from enstools.core import check_arguments
import numpy as np
import xarray
import scipy.spatial
from collections import OrderedDict


class NearestNeighbourInterpolator:
    """
    This class performs the actual interpolation. It is initialised by nearest_neighbour, not directly
    """
    def __init__(self, indices, weights, distances, shape, n_target_points, n_source_points, target_shape, target_lon, target_lat):
        self._indices = indices
        self._weights = weights
        self._distances = distances
        self._shape = shape
        self._n_target_points = n_target_points
        self._n_source_points = n_source_points
        self._target_shape = target_shape
        self._target_lon = target_lon
        self._target_lat = target_lat

    def __call__(self, data):
        """
        Interpolate the data given on the input grid to the points specified during the creation of the interpolator.

        Parameters
        ----------
        data : np.ndarray of xarray.DataArray
                Input array with the same shape as defined by the coordinates used to create the interpolator

        Returns
        -------
        np.ndarray or xarray.DataArray with the shape of the point coordinates used in the creation of the interpolator.
        If the array to interpolate has more dimensions, then these dimensions are prepended to the result, e.g.
        (level, points).
        """
        # check the shape of the data array
        if data.shape[-len(self._shape):] != self._shape:
            raise ValueError("the rightmost dimension of the array to interpolate from must be %s" % (self._shape,))

        # how man dimensions should the result have?
        if len(data.shape) == len(self._shape):
            result = np.empty(self._n_target_points)
        else:
            result = np.empty(data.shape[:-len(self._shape)] + (self._n_target_points,))

        # perform the actual calculation on numpy arrays
        self.__perform_interpolation(np.asarray(data), result)

        # reshape if necessary
        if data.ndim > len(self._shape):
            target_shape = data.shape[:data.ndim-len(self._shape)] + self._target_shape
        else:
            target_shape = self._target_shape
        if result.shape != target_shape:
            result = result.reshape(target_shape)

        # create xarray dataset
        result_coords = OrderedDict()
        result_attrs = OrderedDict()
        result_name = "interpolated"
        if isinstance(data, xarray.DataArray):
            result_attrs = data.attrs
            result_dims = data.dims[:data.ndim-len(self._shape)]
            for one_dim in result_dims:
                result_coords[one_dim] = data.coords[one_dim]
            result_name = data.name
        else:
            if len(result.shape) > 1:
                result_dims = ("dim_0",)
                for d in range(1, data.ndim-len(self._shape)):
                    result_dims += ("dim_%d" % d,)
            else:
                result_dims = ()
        if len(self._target_shape) == 2:
            result_dims += ("lat", "lon")
            result_coords["lat"] = self._target_lat
            result_coords["lon"] = self._target_lon
            result_attrs["grid_type"] = "regular_ll"
        else:
            result_dims += ("cell",)
            result_coords["lat"] = xarray.DataArray(np.asarray(self._target_lat), dims=("cell",))
            result_coords["lon"] = xarray.DataArray(np.asarray(self._target_lon), dims=("cell",))
            result_attrs["grid_type"] = "unstructured_grid"
            result_attrs["coordinates"] = "lon lat"
        result = xarray.DataArray(result, dims=result_dims, coords=result_coords, attrs=result_attrs, name=result_name)
        return result

    def __perform_interpolation(self, data, result):
        """
        perform the actual calculation

        Parameters
        ----------
        data : np.ndarray
        result : np.ndarray
        """
        # TODO: modify for numba jit
        if len(self._shape) == 1:
            if self._n_source_points == 1:
                result[..., :] = data[..., self._indices]
            else:
                for i in range(self._n_target_points):
                    data_values = data[..., self._indices[i]]
                    result[..., i] = np.sum(data_values * self._weights[i, ...], axis=-1)
        elif len(self._shape) == 2:
            if self._n_source_points == 1:
                if self._n_target_points == 1:
                    result[..., 0] = data[..., self._indices[0], self._indices[1]]
                else:
                    for i in range(self._n_target_points):
                        result[..., i] = data[..., self._indices[0][i], self._indices[1][i]]
            else:
                for i in range(self._n_target_points):
                    data_values = data[..., self._indices[0][i], self._indices[1][i]]
                    result[..., i] = np.sum(data_values * self._weights[i, ...], axis=-1)


@check_arguments(units={"src_lon": "degrees_east",
                        "src_lat": "degrees_north",
                        "dst_lon": "degrees_east",
                        "dst_lat": "degrees_north"})
def nearest_neighbour(src_lon, src_lat, dst_lon, dst_lat, src_grid="regular", dst_grid="unstructured", npoints=1, method="mean"):
    """
    Find the coordinates of station locations within gridded model data. Supported are 1d- and 2d-coordinates of regular
    grids (e.g. rotated lat-lon) or 'unstructured' grids like the ICON grid.

    Parameters
    ----------
    src_lon : np.ndarray or xarray.DataArray
            1d or 2d coordinate in x-direction of the source grid

    src_lat : np.ndarray or xarray.DataArray
            1d or 2d coordinate in y-direction of the source grid

    dst_lon : np.ndarray or xarray.DataArray
            1d coordinate in x-direction of the station locations

    dst_lat : np.ndarray or xarray.DataArray
            1d coordinate in y-direction of the station locations

    src_grid : {'regular', 'unstructured'}
            Type of input grid. Possible values are:
            "regular": a regular grid with 1d or 2d coordinates. 1d-coordinates are internally converted to 2d-
            coordinates using meshgrid. This selection is the default.
            "unstructured": The grid is given as a 1d list of points (e.g., station data or ICON model output).

    dst_grid : {'regular', 'unstructured'}
            Type of output grid. Possible values are:
            "regular": a regular grid with 1d coordinates.
            "unstructured": The grid is given as a 1d list of points (e.g., station data or ICON model output). This
            selection is the default.

    npoints : int
            Number of nearest points to be used in the interpolation. For a regular grid useful values are 4 or 12. For
            data on the ICON grid 3 or 6 might be used. The actual geometry of the grid is not taken into account. The
            points are selected based merely on distance.

    method : {'mean', 'd-mean'}
            "mean" : the mean value of the neighbour points is used with equal weight.
            "d-mean": each point is weighted by the reciprocal of the squared distance. The minimum distance within this
            calculation is half of the mean grid spacing.

    Returns
    -------
    NearestNeighbourInterpolator
            callable interpolator object. Each call returns interpolated values

    Examples
    --------
    >>> import numpy
    >>> if numpy.__version__ > "1.13.9": numpy.set_printoptions(legacy="1.13")      # numpy version independent printing
    >>> lon = numpy.arange(10)
    >>> lat = numpy.arange(15)
    >>> gridded_data = numpy.zeros((15, 10))
    >>> gridded_data[8, 4] = 3
    >>> f = nearest_neighbour(lon, lat, 4.4, 7.6)
    >>> f(gridded_data)
    <xarray.DataArray 'interpolated' (cell: 1)>
    array([ 3.])
    Coordinates:
        lat      (cell) float64 7.6
        lon      (cell) float64 4.4
    Dimensions without coordinates: cell
    Attributes:
        grid_type:    unstructured_grid
        coordinates:  lon lat
    """
    # create an array containing all coordinates
    if src_grid == "regular":
        if src_lon.ndim == 1:
            lon_2d, lat_2d = np.meshgrid(src_lon, src_lat)
        elif src_lon.ndim == 2:
            if src_lon.shape != src_lat.shape:
                raise ValueError("for 2d-coordinates, the shapes have to match!")
            lon_2d, lat_2d = src_lon, src_lat
        else:
            raise ValueError("only 1d- and 2d-coordinates are supported for regular grids")
        coords = np.stack((lon_2d.flatten(), lat_2d.flatten()), axis=1)
        input_dims = lon_2d.shape
    elif src_grid == "unstructured":
        if src_lon.ndim == 1:
            coords = np.stack((src_lon, src_lat), axis=1)
            input_dims = src_lon.shape
        else:
            raise ValueError("an unstructured grid is supposed to have 1d-coordinate arrays!")

    # convert point coordinates if given as scalar
    if not hasattr(dst_lon, "__len__"):
        dst_lon = np.array((dst_lon,))
        dst_lat = np.array((dst_lat,))

    # keep the target coordinates
    target_lon = dst_lon
    target_lat = dst_lat

    # create coordinates for regular output grid
    if dst_grid == "regular" and dst_lon.ndim == 1 and dst_lat.ndim == 1:
        point_lon_2d, point_lat_2d = np.meshgrid(dst_lon, dst_lat)
        dst_lon = point_lon_2d.flatten()
        dst_lat = point_lat_2d.flatten()
        target_shape = point_lon_2d.shape
    else:
        target_shape = (len(dst_lon),)

    # create the kd-tree and calculate the indices and the weights for the indices
    kdtree = scipy.spatial.cKDTree(coords)

    # estimate the distance between grid points by calculating the distance to the next neighbour for three points
    grid_point_distance, _ = kdtree.query(coords[[0, len(coords) // 2, len(coords)-1]], k=2)
    mean_grid_point_distance = grid_point_distance[:, 1].mean()

    # calculate indices and distances for the target points
    if len(dst_lon) == 1:
        distances, indices_flat = kdtree.query((dst_lon[0], dst_lat[0]), k=npoints)
    else:
        distances, indices_flat = kdtree.query(np.stack((dst_lon, dst_lat), axis=1), k=npoints)
    if npoints > 1 and len(distances.shape) == 1:
        distances = np.expand_dims(distances, 0)
        indices_flat = np.expand_dims(indices_flat, 0)
    if npoints > 1:
        if method == "mean":
            weights = np.empty(distances.shape)
            weights[:] = 1.0 / npoints
        elif method == "d-mean":
            weights = 1.0 / np.maximum(distances, mean_grid_point_distance / 2.0)**2
            weights = weights / np.tile(np.expand_dims(np.sum(weights, 1), 1), npoints)
        else:
            raise ValueError("unsupported method: %s" % method)
    else:
        weights = 1
    if len(input_dims) > 1:
        indices = np.unravel_index(indices_flat, input_dims)
    else:
        indices = indices_flat
    if type(indices) == int:
        indices = np.asarray([indices])

    # construct and return the interpolator object
    return NearestNeighbourInterpolator(indices, weights, distances, input_dims, len(dst_lon), npoints, target_shape, target_lon, target_lat)



# some domain specific service routines for xarray.Datasets
import six


def drop_unused(ds):
    """
    COSMO is one example of a model which stores all possible coordinates within the output files even if no data
    variables are present which would use these variables. This function checks all coordinates and removes those that
    are not used by any data variable. Unused dimensions are also removed from the dataset.

    Some special coordinates are kept:
    - rotated_pole

    Parameters
    ----------
    ds : xarray.Dataset
            Dataset to check.

    Returns
    -------
    xarray.Dataset:
            a copy of the input dataset with removed coordinates.
    """
    # list of unremovable coordinates
    unremovable = ["rotated_pole"]

    # loop over all coordinates
    coords_to_remove = []
    for one_coord_name, one_coord_var in six.iteritems(ds.coords):
        # skip unremoveable coordinates
        if one_coord_name in unremovable:
            continue

        # this coordinate is only kept if all its dimensions are used in any of the data variables
        has_all_dims_in_any_var = False
        for one_name, one_var in six.iteritems(ds.data_vars):
            has_all_dims = len(one_coord_var.dims) > 0
            for one_dim in one_coord_var.dims:
                if not one_dim in one_var.dims:
                    has_all_dims = False
                    break
            if has_all_dims:
                has_all_dims_in_any_var = True
                break

        # mark this coord for deletion?
        if not has_all_dims_in_any_var:
            coords_to_remove.append(one_coord_name)

    # actually remove the coords
    if len(coords_to_remove) > 0:
        for one_coord_name in coords_to_remove:
            if one_coord_name in ds.indexes:
                ds = ds.reset_index(one_coord_name, drop=True)
            elif one_coord_name in ds.coords:
                ds = ds.reset_coords(one_coord_name, drop=True)

    # remove unused dimensions
    unused_dims = []
    for one_dim in ds.dims:
        is_used = False
        for one_name, one_var in six.iteritems(ds.data_vars):
            if one_dim in one_var.dims:
                is_used = True
                break
        if not is_used:
            unused_dims.append(one_dim)

    # actually remove unused dims. This is not straight forward: it is necessary to create a dummy variable. That will
    # trigger recalculation of the dimensions.
    if len(unused_dims) > 0:
        ds["dummy_for_triggering_dims_calculation"] = 0
        del ds["dummy_for_triggering_dims_calculation"]

    return ds


import xarray
import numpy
from scipy.stats import pearsonr


def pearsonr_wrapper(a: numpy.ndarray, b: numpy.ndarray) -> float:
    """
    Small wrapper for the pearsonr function, converting the input arrays to 1D arrays
    """
    corr, _ = pearsonr(a.ravel(), b.ravel())
    return corr


def pearson_correlation(reference: xarray.DataArray, target: xarray.DataArray) -> xarray.DataArray:
    """
    Description:     It returns the Pearson's correlation. Uses implementation from scipy.

    Functions
    :param reference:
    :param target:
    :return:
    """
    import warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

    core_dims = [d for d in reference.dims if d != "time"]

    # Load data
    reference.load()
    target.load()

    return xarray.apply_ufunc(pearsonr_wrapper, reference, target,
                              #
                              vectorize=True,
                              input_core_dims=[core_dims, core_dims],
                              output_core_dims=[[]],
                              )


def pearson_correlation_index(reference: xarray.DataArray, target: xarray.DataArray) -> xarray.DataArray:
    """
    Description:
        It returns the Pearson's correlation index. Uses implementation from scipy.
        This metrics is expected to be useful when te values are close to 1.

    Functions
    :param reference:
    :param target:
    :return:
    """
    correlation = pearson_correlation(reference, target)

    # The function pearson_correlation from scipy returns a NaN if one of the input arrays is constant
    # which can happen quite often if the compression is extreme.
    # We'll instead return a correlation of 0.
    correlation = correlation.fillna(0)

    return xarray.where(correlation == 1.0, numpy.inf, - numpy.log10(1-correlation))




import numpy as np
import xarray
import dask.array as da


def __normalize(variable):
    """
    Scale a variable to mean zero and standard deviation 1

    Parameters
    ----------
    variable : xarray.DataArray or np.ndarray

    Returns
    -------
    xarray.DataArray or np.ndarray
    """
    mean = variable.mean()
    std = variable.std()
    if std != 0:
        result = (variable - mean) / std
    else:
        result = variable - mean
    return result


def prepare(*variables, **kwargs):
    """

    Parameters
    ----------
    variables : xarray.DataArray or np.ndarray
            one or more variables which should be used for the clustering.

    **kwargs
            *ens_dim* : int or string
                index or name of the dimension along which the clustering should be performed. If not provided,
                xarrays will be scanned for standard ensemble dimension names (ens, ensemble, member, members). If
                no standard ensemble dimension is found, dimension 0 is used. That is also the default for numpy arrays.

    Returns
    -------
    np.ndarray
            2d-array with dimensions (ensemble, feature) where ensemble is the dimension along
            which the clustering should be performed and feature is the product of grid points and number of
            variables.
    """

    # 1. step: check the shape of all input variables
    shape = variables[0].shape
    for ivar, one_var in enumerate(variables):
        if one_var.shape != shape:
            raise ValueError("all input variables have to have the same shape! First variable: %s, variable %d: %s"
                             % (shape, ivar+1, one_var.shape))

    # 2. calculate the shape of the result
    ens_dim = kwargs.get("ens_dim", None)
    if ens_dim is None:
        if isinstance(variables[0], xarray.DataArray):
            for idim, dim_name in enumerate(variables[0].dims):
                if dim_name in ["ens", "ensemble", "member", "members"]:
                    ens_dim = idim
                    break
        if ens_dim is None:
            ens_dim = 0
    if isinstance(ens_dim, str):
        if isinstance(variables[0], xarray.DataArray):
            ens_dim = variables[0].dims.index(ens_dim)
        else:
            raise ValueError("dimension names are only supported for xarray.DataArray variables!")
    if ens_dim > len(shape)-1:
        raise ValueError("the specified ensemble dimension is larger the the number dimension id the input array!")

    feature_shape = list(shape)
    del feature_shape[ens_dim]
    feature_shape = (np.product(feature_shape),)

    # 3. flatten all arrays
    var_list = []
    for one_var in variables:
        # is it an xarray? use directly the data array
        if isinstance(one_var, xarray.DataArray):
            one_var = one_var.data
        # reorder to bring the ensemble dimension to the front
        if ens_dim != 0:
            one_var = np.moveaxis(one_var, ens_dim, 0)
        # flatten all other dimensions
        if one_var.ndim > 2:
            one_var = np.reshape(one_var, (one_var.shape[0],) + feature_shape)
        var_list.append(one_var)

    # 4. normalize all arrays
    var_list = list(map(lambda x: __normalize(x), var_list))

    # Finally concatenate all variables
    return da.concatenate(var_list, axis=1)



import tempfile
import shutil
import atexit
import os


class TempDir(object):
    """
    This class represents a temporal directory which is automatically deleted on exit
    """

    def __init__(self, parentdir=None, check_free_space=True, cleanup=True):
        """
        Parameters
        ----------
        parentdir : str
                the directory is created in the parent directory. If the argument is none, then the directory is created in the default temp-Directory

        check_free_space : bool
                if true, at least 10GB free space in the temporal directory if necessary

        cleanup: bool
                if True, the folder will be deleted at exit of python.
        """
        # create the directory
        if parentdir is not None and not os.path.exists(parentdir):
            try:
                os.makedirs(parentdir)
            except:
                raise IOError("creating temporal folder in %s failed!" % parentdir)
        self.__path = tempfile.mkdtemp(dir=parentdir)

        # check the free disk space
        if check_free_space:
            stat = os.statvfs(self.__path)
            if stat.f_frsize * stat.f_bavail < 1e10:
                raise IOError("""less than 10GB disk space for temporal directory!
                                 set one of the environment variables TMPDIR, TEMP, or TMP to specify a location with more space available!""")

        # register the cleanup-method
        if cleanup:
            atexit.register(self.cleanup)

    def cleanup(self):
        """
        Remove the temporal directory and its content.
        This method is automatically called on exit
        """
        # check is the directory is still there
        if not os.path.exists(self.__path):
            return

        # delete the directory
        shutil.rmtree(self.__path)

    def getpath(self):
        """
        get the path of the temporal directory
        """
        return self.__path



"""
Core functionality used by other components of the ensemble tools
"""
import six
import sys
import re
import logging
import pint
import inspect
import xarray
import numpy
import dask.array
import dask.multiprocessing
import distributed
import string
from decorator import decorator
from pint import DimensionalityError
from .cluster import init_cluster, get_num_available_procs, get_client_and_worker, all_workers_are_local, \
    RoundRobinWorkerIterator
from .os_support import getstatusoutput, get_cache_dir

# to convert enstools into a namespace package, the version is now listed here and not in the level above
__version__ = "2022.9.4"


class UnitRegistry(pint.UnitRegistry):
    # workaround for Pint issue: https://github.com/hgrecco/pint/issues/476
    def __getattr__(self, name):
        if name[0] == '_':
            try:
                value = super(UnitRegistry, self).__getattr__(name)
                return value
            except pint.errors.UndefinedUnitError as e:
                raise AttributeError()
        else:
            return super(UnitRegistry, self).__getattr__(name)

    # adapted parser for units with minus sign
    def __call__(self, *args, **kwargs):
        return super(UnitRegistry, self).__call__(re.sub("([a-zA-Z]+)(-[0-9]+)", r"\g<1>**\g<2>", args[0]))


# do not use the cache with dask 0.18.0: https://github.com/dask/dask/pull/3632
if dask.__version__ != "0.18.0":
    from dask.cache import Cache
    cache = Cache(2e9)
    cache.register()

# all units from pint
ureg = UnitRegistry()
# add specific units
ureg.define("degrees_east = deg = degree_east = degree_E = degrees_E = degreeE = degreesE.")
ureg.define("degrees_north = deg = degree_north = degree_N = degrees_N = degreeN = degreesN")

# default settings
__default_settings = {"check_arguments:convert": True,
                      "check_arguments:strict": False,
                      "check_arguments:reorder": True}

# default style for logging
logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")


def set_behavior(check_arguments_convert=None, check_arguments_strict=None, check_arguments_reorder=None, log_level=None):
    """
    Change the default behavior of the @check_arguments decorator

    Parameters
    ----------
    check_arguments_convert : bool
            if true, automatic unit conversion takes place

    check_arguments_strict : bool
            if true, missing unit or dimension information cause an exception

    """
    if check_arguments_convert is not None:
        __default_settings["check_arguments:convert"] = check_arguments_convert
    if check_arguments_strict is not None:
        __default_settings["check_arguments:strict"] = check_arguments_strict
    if check_arguments_reorder is not None:
        __default_settings["check_arguments:reorder"] = check_arguments_reorder
    if log_level is not None:
        if log_level not in ["ERROR", "WARN", "DEBUG", "INFO"]:
            raise ValueError("unsupported log level: '%s'" % log_level)
        __set_log_level(log_level)
        # set the log level also on all workers
        try:
            client = distributed.get_client()
            client.run(__set_log_level, log_level)
        except:
            pass


def __set_log_level(log_level):
    """
    a function executable on all worker processes

    Parameters
    ----------
    level : str
            new log level
    """
    logging.getLogger().setLevel(log_level)


def __replace_argument(args, index, new_arg):
    """
    Replace one argument from within an argument tuple

    Parameters
    ----------
    args : tuple
            tuple of arguments
    index : int
            position of the argument within the tuple
    new_arg
            value of the new argument

    Returns
    -------
    tuple
            new tuple of arguments
    """
    arg_list = list(args)
    arg_list[index] = new_arg
    return tuple(arg_list)


def __shapes_are_equal(shape1, shape2, named_dim_length):
    """
    compare the shape of two array. zeros in the second arguments are ignored

    Parameters
    ----------
    shape1 : tuple
    shape2 : tuple
    named_dim_length : dict

    Returns
    -------
    bool
            True, if both shapes are identical
    """
    if len(shape1) != len(shape2):
        return False
    for i in range(len(shape2)):
        if type(shape2[i]) == str:
            # this and all following dimensions have to match
            if shape2[i] in named_dim_length:
                shape2_entry = named_dim_length[shape2[i]]
            else:
                if shape2[i].endswith(":"):
                    named_dim_length[shape2[i]] = shape1[i:]
                else:
                    named_dim_length[shape2[i]] = shape1[i]
                continue
        else:
            shape2_entry = shape2[i]
        if type(shape2_entry) == tuple:
            if shape1[i:] != shape2_entry:
                return False
        elif shape1[i] != shape2_entry and shape2_entry != 0:
            return False
    return True


def get_arg_spec(func):
    """
    returns the arguments of a function and works for python 2 and 3.

    Parameters
    ----------
    func: callable
            function object to inspect
    """
    if six.PY3:
        arg_spec = inspect.getfullargspec(func)
    else:
        arg_spec = inspect.getargspec(func)
    return arg_spec


def check_arguments(units={}, dims={}, shape={}):
    """
    Parameters
    ----------
    units : dict
            Definition of the argument units in the format {"argument_name|argument_index": "unit"}

    dims : dict
            Definition of the argument dimensions in the format {"argument_name|argument_index": ("time","lon","lat")}
            for an exact ordering of the dimensions or {"argument_name|argument_index": ["time","lon","lat"]} for
            arbitrary ordering of dimensions. In the later case, the number of dimensions is not checked, which means
            that the actual argument is allowed to have more dimensions as long as the mentioned dimensions are present.

    shape: dict
            Definition of the arguments shape in the format {"argument_name|argument_index": (10,17)} for a fixed shape
            or {"argument_name|argument_index": "other_argument_name"} if two arguments are required to have the same
            shape. In the first case, zero may be used to indicate, that the size in a direction does not matter, a
            letter may be used if two arguments should have the same shape in this direction. A colon my be added to the
            letter to indicate, that all following dimension have to be identical.

    Returns
    -------

    """

    @decorator
    def check_arguments_decorator(func, *args, **kwargs):
        # check the actual values of the functions arguments
        arg_spec = get_arg_spec(func)
        arg_names = arg_spec[0]
        arg_values = inspect.getcallargs(func, *args, **kwargs)
        # create a list of the new arguments
        modified_args = []
        # dictionary for named dimension length.
        named_dim_length = {}

        # loop over all none-keyword arguments
        for iarg, one_arg_name in enumerate(arg_names):
            current_argument = arg_values[one_arg_name]

            # check the units
            if one_arg_name in units or iarg in units:
                # construct the unit for this argument
                if iarg in units:
                    target_arg_unit = ureg(units[iarg])
                else:
                    target_arg_unit = ureg(units[one_arg_name])

                # is there a units attribute on the variable? Only xarray.DataArrays can have one
                if isinstance(current_argument, xarray.DataArray) and "units" in current_argument.attrs:
                    actual_arg_unit = ureg(current_argument.attrs["units"])

                    # the units differ? try to find a conversion!
                    if target_arg_unit != actual_arg_unit:
                        if __default_settings["check_arguments:convert"]:
                            conversion_failed = False
                            try:
                                factor = actual_arg_unit.to(target_arg_unit)
                                current_argument = current_argument * factor.magnitude
                                logging.warning("The unit of the argument '%s' was converted from '%s' to '%s' by multiplication with the factor %f" % (arg_names[iarg], actual_arg_unit, target_arg_unit, factor.magnitude))
                            except DimensionalityError as ex:
                                conversion_failed = True
                                error_msg = "%s; Unable to convert units of argument '%s' of method '%s'!" % (str(ex), arg_names[iarg], func.__name__)
                            if conversion_failed:
                                raise ValueError(error_msg)
                        else:
                            logging.warning("The unit of the argument '%s' differs from '%s', no conversion was done!" % (arg_names[iarg], target_arg_unit))

                # no units attribute, is that an error?
                else:
                    # construct error or warning message
                    msg = "Argument '%s' has no unit information in form of the 'units' attribute" % arg_names[iarg]
                    if __default_settings["check_arguments:convert"] and __default_settings["check_arguments:strict"]:
                        raise ValueError(msg)
                    else:
                        logging.warning(msg + ", assuming '%s'" % target_arg_unit)

            # check the dimensions of the argument
            if one_arg_name in dims or iarg in dims:
                # is it an xarray?
                if isinstance(current_argument, xarray.DataArray):
                    actual_arg_dims = current_argument.dims
                    if iarg in dims:
                        target_arg_dims = dims[iarg]
                    else:
                        target_arg_dims = dims[one_arg_name]

                    # are the target dimensions exact or are only required dimensions given?
                    if type(target_arg_dims) == tuple:
                        if actual_arg_dims != target_arg_dims:
                            if __default_settings["check_arguments:reorder"]:
                                # dimensions differ, is it possible to solve the problem by reordering?
                                if len(actual_arg_dims) == len(target_arg_dims) and sorted(actual_arg_dims) == sorted(target_arg_dims):
                                    current_argument = current_argument.transpose(*target_arg_dims)
                                    logging.warning("The dimensions of the argument '%s' were reordered from %s to %s" % (arg_names[iarg], actual_arg_dims, target_arg_dims))
                                else:
                                    raise ValueError("unable to change dimensions %s automatically to %s" % (actual_arg_dims, target_arg_dims))
                            else:
                                logging.warning(
                                    "The dimensions of the argument '%s' differs from %s, no reordering was done!" % (
                                    arg_names[iarg], target_arg_dims))

                    # dims are given, but their ordering does not matter
                    elif type(target_arg_dims) == list:
                        for one_dim in target_arg_dims:
                            if one_dim not in actual_arg_dims:
                                msg = "Argument '%s' has no dimension '%s'" % (arg_names[iarg], one_dim)
                                if __default_settings["check_arguments:strict"]:
                                    raise ValueError(msg)
                                else:
                                    logging.warning(msg)

                # no xarray, unable to check dimensions
                else:
                    msg = "Argument '%s' has no dimension name information!" % arg_names[iarg]
                    if __default_settings["check_arguments:reorder"] and __default_settings["check_arguments:strict"]:
                        raise ValueError(msg)
                    else:
                        logging.warning(msg)

            # check the shape of the argument
            if (one_arg_name in shape or iarg in shape) and hasattr(current_argument, "shape"):
                if iarg in shape:
                    shape_entry = shape[iarg]
                else:
                    shape_entry = shape[one_arg_name]
                # is the shape given as tuple or as name of another variable?
                if type(shape_entry) == tuple:
                    if not __shapes_are_equal(current_argument.shape, shape_entry, named_dim_length):
                        raise ValueError("The shape of the argument '%s', which is %s, differs from the pre-defined shape %s" % (one_arg_name, current_argument.shape, shape_entry))
                # shape should be identical to other variables shape
                else:
                    if shape_entry in arg_names:
                        if hasattr(arg_values[shape_entry], "shape"):
                            target_shape = arg_values[shape_entry].shape
                            if current_argument.shape != target_shape:
                                raise ValueError(
                                    "The shape of the argument '%s', which is %s, differs from the shape %s of the reference variable '%s'!" % (
                                    one_arg_name, current_argument.shape, target_shape, shape_entry))
                        else:
                            raise ValueError("The reference variable '%s' has no shape attribute!" % shape_entry)

            # construct new argument list
            if iarg > len(modified_args) and one_arg_name in kwargs:
                kwargs[one_arg_name] = current_argument
            else:
                modified_args.append(current_argument)
                if one_arg_name in kwargs:
                    del kwargs[one_arg_name]

        # perform the actual calculation
        modified_args = tuple(modified_args)
        return_value = func(*modified_args, **kwargs)

        # check the units and convert if necessary and requested
        if "return_value" in units:
            if isinstance(return_value, xarray.DataArray) and "units" in return_value.attrs:
                # get actual and target units and compare
                actual_return_unit = ureg(return_value.attrs["units"])
                target_return_unit = ureg(units["return_value"])
                if actual_return_unit != target_return_unit:
                    if __default_settings["check_arguments:convert"]:
                        try:
                            factor = actual_return_unit.to(target_return_unit)
                            return_value = return_value * factor.magnitude
                            return_value.attrs["units"] = units["return_value"]
                            logging.warning(
                                "The unit of the return value of method '%s' was converted from '%s' to '%s' by multiplication with the factor %f" % (
                                func.__name__, actual_return_unit, target_return_unit, factor.magnitude))
                        except DimensionalityError as ex:
                            ex.extra_msg = "; Unable to convert units of return value of method '%s'!" % (func.__name__)
                            raise

            # unable to check!
            else:
                if __default_settings["check_arguments:convert"] or __default_settings["check_arguments:strict"]:
                    msg = "Return value of function '%s' has no unit information in form of the 'units' attribute" % func.__name__
                    raise ValueError(msg)

        # check the dimensions and convert if necessary and requested
        if "return_value" in dims:
            # is it an xarray?
            if isinstance(return_value, xarray.DataArray):
                actual_arg_dims = return_value.dims
                target_arg_dims = dims[one_arg_name]

                # are the target dimensions exact or are only required dimensions given?
                if type(target_arg_dims) == tuple:
                    if actual_arg_dims != target_arg_dims:
                        if __default_settings["check_arguments:reorder"]:
                            # dimensions differ, is it possible to solve the problem by reordering?
                            if len(actual_arg_dims) == len(target_arg_dims) and sorted(actual_arg_dims) == sorted(
                                    target_arg_dims):
                                return_value = return_value.transpose(*target_arg_dims)
                                logging.warning(
                                    "The dimensions of the return value were reordered from %s to %s" % (actual_arg_dims, target_arg_dims))
                            else:
                                raise ValueError("unable to change dimensions of return value %s automatically to %s" % (actual_arg_dims, target_arg_dims))
                        else:
                            logging.warning(
                                "The dimensions of the return value differs from %s, no reordering was done!" % (
                                    arg_names[iarg], target_arg_dims))

                # dims are given, but their ordering does not matter
                elif type(target_arg_dims) == list:
                    for one_dim in target_arg_dims:
                        if one_dim not in actual_arg_dims:
                            msg = "The return value has no dimension '%s'" % one_dim
                            if __default_settings["check_arguments:strict"]:
                                raise ValueError(msg)
                            else:
                                logging.warning(msg)

            # no xarray, unable to check dimensions
            else:
                msg = "The return value has no dimension name information!"
                if __default_settings["check_arguments:reorder"] or __default_settings["check_arguments:strict"]:
                    raise ValueError(msg)
                else:
                    logging.warning(msg)

        # check the shape of the return value
        if "return_value" in shape:
            # is the shape given as tuple or as name of another variable?
            if type(shape["return_value"]) == tuple:
                if not __shapes_are_equal(return_value.shape, shape["return_value"], named_dim_length):
                    raise ValueError(
                        "The shape of the return value, which is %s, differs from the pre-defined shape %s" % (
                        return_value.shape, shape["return_value"]))
            # shape should be identical to other variables shape
            else:
                if shape["return_value"] in arg_names:
                    if hasattr(arg_values[shape["return_value"]], "shape"):
                        target_shape = arg_values[shape["return_value"]].shape
                        if return_value.shape != target_shape:
                            raise ValueError(
                                "The shape of the return value, which is %s, differs from the shape %s of the reference variable '%s'!" % (
                                    return_value.shape, target_shape, shape["return_value"]))
                    else:
                        raise ValueError(
                            "The reference variable '%s' has no shape attribute!" % shape["return_value"])

        # finally return the result
        return return_value

    return check_arguments_decorator


def get_chunk_size_for_n_procs(shape, nproc):
    """
    Create a chunk distribution for dask-arrays depending on the number of available processors. Chunks are created
    from the two right-most dimensions.

    Parameters
    ----------
    nproc : int
            number of available processors

    shape : tuple
            shape of the array to split up

    Returns
    -------
    tuple
            chunk-tuple usable in dask arrays

    Examples
    --------
    >>> get_chunk_size_for_n_procs(100, 10)
    (10,)
    >>> get_chunk_size_for_n_procs((100, 100), 8)
    (50, 25)
    >>> get_chunk_size_for_n_procs((1000, 100), 8)
    (125, 100)
    >>> get_chunk_size_for_n_procs((100, 1000), 8)
    (100, 125)
    >>> get_chunk_size_for_n_procs((10, 1000, 1000), 8)
    (10, 500, 250)
    """
    if type(shape) == int:
        shape = (shape,)
    # use the right-most two dims
    if len(shape) <= 2:
        shape_for_chunk = shape
    else:
        shape_for_chunk = shape[-2:]

    # decide for the number of chunks in each direction
    if len(shape_for_chunk) == 1:
        return (shape_for_chunk[0] // nproc,)
    if len(shape_for_chunk) == 2:
        ratio = shape_for_chunk[1] / float(shape_for_chunk[0])
        if 0.5 <= ratio <= 2.0:
            n_chunks_0 = int(numpy.sqrt(nproc))
            n_chunks_1 = nproc // n_chunks_0
        elif ratio < 0.5:
            n_chunks_0 = nproc
            n_chunks_1 = 1
        elif ratio > 2.0:
            n_chunks_0 = 1
            n_chunks_1 = nproc
        if len(shape) > len(shape_for_chunk):
            return shape[:-len(shape_for_chunk)] + (shape_for_chunk[0] // n_chunks_0, shape_for_chunk[1] // n_chunks_1)
        else:
            return shape_for_chunk[0] // n_chunks_0, shape_for_chunk[1] // n_chunks_1


def parallelize_univariate_two_arg(func):
    """
    Parallelize a function with two arguments. The first argument is a (N,...)-array, the second argument is a
    (e,N,...)-array. The array in the first argument will be divided into chunks of (almost) equal size. The array in
    the second argument will be divided in the same way in its rightmost dimensions.

    The first call to the underlining function will be func(arg0[0:chunk0], arg1[:,0:chunk0]),
    the second func(arg0[chunk0:chunk1], arg1[:,chunk0:chunk1]), ...

    Parameters
    ----------
    func : function object

    Returns
    -------
    numpy.ndarray or xarray.DataArray or float
    """

    # create a wrapper for argument and result conversion conversion
    @check_arguments(shape={0: ("n-obs:",), 1: ("ensemble", "n-obs:")})
    def function_wrapper(arg0, arg1, mean=False, **kwargs):
        # paralyze with dask
        # calculate chunk sizes
        nprocs = get_num_available_procs()
        chunk_size = get_chunk_size_for_n_procs(arg0.shape, nprocs)
        da0 = dask.array.from_array(arg0, chunks=chunk_size)
        da1 = dask.array.from_array(arg1, chunks=(arg1.shape[0],) + chunk_size)

        # perform the actual calculation on a chunk of the array
        def dask_calculation(a0, a1):
            if len(a0.shape) > 1:
                original_shape = a0.shape
                a0 = a0.flatten()
                a1 = a1.reshape((a1.size // a0.size, a0.size))
                result = func(a0, numpy.moveaxis(a1, 0, -1), **kwargs)
                result = result.reshape(original_shape)
            else:
                result = func(a0, numpy.moveaxis(a1, 0, -1), **kwargs)
            return result

        # order of index in input and output
        obs_ind = string.ascii_lowercase[:len(chunk_size)]
        fct_int = "z"+obs_ind

        # perform the actual calculation
        result = dask.array.atop(dask_calculation, obs_ind, da0, obs_ind, da1, fct_int, dtype=da0.dtype, concatenate=True)

        # calculate mean if requested
        if mean:
            return result.mean().compute(get=dask.multiprocessing.get, num_workers=nprocs)
        else:
            return result.compute(get=dask.multiprocessing.get, num_workers=nprocs)

    return function_wrapper


def vectorize_univariate_two_arg(func):
    """
    Vectorize a function with two arguments. The first argument if a 1d-array, the second argument is a 2d-array. The
    first call to the underlining function will be func(arg0[0], arg1[:,0]), the second func(arg0[1], arg1[:,1]), ...

    Parameters
    ----------
    func : function object

    Returns
    -------
    numpy.ndarray or xarray.DataArray or float
    """

    # create the vectorized version
    vfunc = numpy.vectorize(func, signature="(),(n)->()")

    # create a wrapper for argument and result conversion conversion
    @check_arguments(shape={0: ("n-obs:",), 1: ("ensemble", "n-obs:")})
    def function_wrapper(arg0, arg1, mean=False, **kwargs):
        # perform the actual calculation
        result = vfunc(arg0, numpy.moveaxis(arg1, 0, -1), **kwargs)

        # calculate mean if requested
        if mean:
            return result.mean()
        else:
            return result

    return function_wrapper


def vectorize_multivariate_two_arg(func, arrays_concatenated=True):
    """
    Vectorize a function with two arguments. The first argument if a 2d-array, the second argument is a 3d-array. The
    first call to the underlining function will be func(arg0[:,0], arg1[:,:,0]), the second
    func(arg0[:,1], arg1[:,:,1]), ...

    Parameters
    ----------
    func : function object
    arrays_concatenated : bool
            if true, the function has one observation and one forecast argument, if false the function has two
            observation and two forecast arguments

    Returns
    -------
    numpy.ndarray or xarray.DataArray or float
    """
    # check the arguments of the functions. only the first two arguments are used in vectorization.
    if sys.version_info >= (3, 0):
        arg_spec = inspect.getfullargspec(func)
    else:
        arg_spec = inspect.getargspec(func)

    # create the vectorized version
    if len(arg_spec.args) > 2:
        vfunc = numpy.vectorize(func, signature="(d),(d,m)->()", excluded=arg_spec.args[2:])
    else:
        vfunc = numpy.vectorize(func, signature="(d),(d,m)->()")

    def __vectorize_multivariate_two_arg_concatenated(arg0, arg1, mean=False, **kwargs):
        # perform the actual calculation
        if type(arg0) == xarray.DataArray:
            arg0 = arg0.data
        if type(arg1) == xarray.DataArray:
            arg1 = arg1.data
        result = vfunc(numpy.moveaxis(arg0, 0, -1), numpy.moveaxis(arg1, (0, 1), (-2, -1)), **kwargs)

        # calculate mean if requested
        if mean:
            return result.mean()
        else:
            return result

    # create a wrapper for argument and result conversion conversion
    if arrays_concatenated:
        return check_arguments(shape={0: ("obs-dim", "n-obs:"), 1: ("obs-dim", "ensemble", "n-obs:")})(__vectorize_multivariate_two_arg_concatenated)

    else:
        @check_arguments(shape={0: ("n-obs:",), 1: ("n-obs:",), 2: ("ensemble", "n-obs:"), 3: ("ensemble", "n-obs:")})
        def function_wrapper(obs0, obs1, fct0, fct1, mean=False, **kwargs):
            # concatenate arrays
            arg0 = numpy.stack((obs0, obs1))
            arg1 = numpy.stack((fct0, fct1))

            # call the version of this function for concatenated arrays
            return __vectorize_multivariate_two_arg_concatenated(arg0, arg1, mean, **kwargs)

        return function_wrapper


def import_multipledispatch(dispatcher, globals):
    """
    Import all implementations of a function created by the multipledispatch module into the globals dictionary.
    They will be made available by name_md%02d, where *%02d* is a counter.

    The purpose of this function is to provide compatibility for multiple dispatch functions with sphinx.

    Parameters
    ----------
    dispatcher
            the dispatcher function object

    globals
            namespace into which the function and all its implementations should be imported

    """
    # iterate over all implementations
    imported = set()
    i = 0
    for signature in sorted(dispatcher.funcs.keys(), key=lambda x: str(x)):
        func = dispatcher.funcs[signature]
        if not func in imported:
            i += 1
            globals["%s_md%02d" % (dispatcher.__name__, i)] = func
            imported.add(func)



import xarray


def positivity(reference: xarray.DataArray, target: xarray.DataArray) -> xarray.DataArray:
    """
    Description:
    This metric evaluates if reference values are strictly positive and checks whether this is
    maintained in the target dataset.
    It returns 1.0 if the reference and the target have the same positivity and -1.0 otherwise.
    These values are arbitrary.


    Functions
    :param reference:
    :param target:
    :return:
    """
    if (reference >= 0.0).all() == (target >= 0.0).all():
        return xarray.DataArray(1.0)
    else:
        return xarray.DataArray(-1.0)




"""
preliminary minimal support for grib files. This module will be replaced as soon as the official eccodes xarray-
support is available.
"""
import logging
try:
    from . import eccodes_cffi
except OSError:
    logging.warning("eccodes c-library not found, grib file support not available!")
    pass
from collections import OrderedDict
import os
import io
import re
import xarray
import dask.array
import numpy
from datetime import datetime, timedelta
import logging
import distributed
from enstools.core import all_workers_are_local


def read_grib_file(filename, debug=False, in_memory=False, leadtime_from_filename=False, client=None, worker=None):
    """
    Read the contents of a grib1 or grib2 file

    Parameters
    ----------
    filename : string
            name of the file to read

    debug :  bool
            show some additional debug messages

    in_memory : bool
            load the data directly into memory

    leadtime_from_filename : bool
            COSMO-GRIB1-Files do not contain exact times. If this argument is set, then the timestamp is calculated
            from the init time and the lead time from the file name.

    client : distributed.client
            dask-distributed Client object used for persisting data into cluster memory

    worker : distributed.worker
            dask-distributed Worker object. This is supposed to be not None when the routine is called inside of a
            worker process.

    Returns
    -------
    xarray.Dataset
            a netcdf-like representation of the file-content
    """
    if "eccodes_cffi" not in globals():
        raise ImportError("eccodes interface not found, grib file support not available!")

    # use always the absolute path, the relative path may change during the lifetime of the dataset
    filename = os.path.abspath(filename)

    # lists of dimensions coordinates and variables
    variables = set()
    levels = {}
    level_values = {}
    dimensions = {}
    dimension_names = {}
    coordinates = OrderedDict()
    attributes = {}
    encodings = {}
    ensemble_members = set()
    times = set()
    datatype = {}
    msg_by_var_level_ens = {}
    rotated_pole = {}

    # list of skipped grid types
    skipped_grids = set()

    # The full message payload is only read if we want data in memory. In a dask cluster,
    # workers take over the data loading. Exception: if all workers are local, load the data directly.
    load_payload = in_memory and client is None
    if client is not None:
        all_local = all_workers_are_local(client)
    else:
        all_local = False
    if in_memory and client is not None and all_local:
        load_payload = True

    # loop to select all messages
    logging.debug("start reading all grib messages from %s ..." % filename)
    gfile = io.open(filename, "rb")
    while True:
        # read the content of the next grib message from the input file.
        offset = gfile.tell()

        # create an empty message from the message raw data.
        msg = eccodes_cffi.GribMessage(gfile, offset, read_data=load_payload)
        if not msg.is_valid():
            break

        # skip messages on unsupported grids
        if msg["gridType"] not in ["sh", "regular_ll", "rotated_ll", "reduced_gg", "unstructured_grid"]:
            if msg["gridDefinitionDescription"] not in skipped_grids:
                skipped_grids.add(msg["gridDefinitionDescription"])
                logging.warning("skipping grib message due to unsupported grid: %s" % msg["gridDefinitionDescription"])
            continue

        # record the name
        variable_id = (msg.get_name(prefer_cf=False), msg["typeOfLevel"])
        if variable_id not in variables:
            variables.add(variable_id)

        # make a list of all levels
        if variable_id not in levels:
            levels[variable_id] = [msg["level"]]
            level_values[variable_id] = {}
        else:
            if msg["level"] not in levels[variable_id]:
                levels[variable_id].append(msg["level"])
        if msg["level"] not in level_values[variable_id]:
            level_values[variable_id][msg["level"]] = msg.get_level()

        # record the horizontal dimensions
        if variable_id not in dimensions:
            dimensions[variable_id], dimension_names[variable_id] = msg.get_dimension(dimensions, dimension_names)

        # record the valid time stamp
        if leadtime_from_filename:
            leadtime = re.search("lfff(\d\d)(\d\d)(\d\d)(\d\d)", os.path.basename(filename))
            if leadtime is None:
                raise IOError("unable to read leadtime from filename: %s", filename)
            initDate = "%08d%04d" % (msg["dataDate"], int(msg["dataTime"]))
            if initDate.startswith("0000"):
                initDate = "2" + initDate[1:]
            time_stamp = datetime.strptime(initDate, "%Y%m%d%H%M")
            time_stamp += timedelta(days=int(leadtime.group(1)), hours=int(leadtime.group(2)), minutes=int(leadtime.group(3)), seconds=int(leadtime.group(4)))
        else:
            validityDate = "%08d%04d" % (msg["validityDate"], int(msg["validityTime"]))
            if validityDate.startswith("0000"):
                validityDate = "2" + validityDate[1:]
            time_stamp = datetime.strptime(validityDate, "%Y%m%d%H%M")
        if time_stamp not in times:
            times.add(time_stamp)

        # horizontal coordinates for unstaggered variables
        if "lon" not in coordinates and not "srlon" in dimension_names[variable_id] and not "srlat" in dimension_names[variable_id]:
            coord_lon, coord_lat = msg.get_coordinates(dimension_names[variable_id])
            if coord_lon is not None and coord_lat is not None:
                coordinates["lon"], coordinates["lat"] = coord_lon, coord_lat

        # for staggered vartiable add slonu/v and slatu/v
        if dimension_names[variable_id][0] == "rlat" and dimension_names[variable_id][1] == "srlon" and "slonu" not in coordinates:
            coord_lon, coord_lat = msg.get_coordinates(dimension_names[variable_id])
            if coord_lon is not None and coord_lat is not None:
                coordinates["slonu"], coordinates["slatu"] = coord_lon, coord_lat
        if dimension_names[variable_id][0] == "srlat" and dimension_names[variable_id][1] == "rlon" and "slonv" not in coordinates:
            coord_lon, coord_lat = msg.get_coordinates(dimension_names[variable_id])
            if coord_lon is not None and coord_lat is not None:
                coordinates["slonv"], coordinates["slatv"] = coord_lon, coord_lat

        # create a list of all ensemble members
        for ensemble_member_key in ["localActualNumberOfEnsembleNumber", "perturbationNumber"]:
            if ensemble_member_key in msg:
                ensemble_member = msg[ensemble_member_key]
                if msg[ensemble_member_key] not in ensemble_members:
                    ensemble_members.add(msg[ensemble_member_key])
                break
            else:
                ensemble_member = -1

        # select a datatype based on the number of bits per value in the grib message
        if variable_id not in datatype:
            if msg["bitsPerValue"] > 32:
                datatype[variable_id] = numpy.float64
            else:
                datatype[variable_id] = numpy.float32

        # collect attributes of this variables
        if variable_id not in attributes:
            attrs = OrderedDict()
            attrs["units"] = msg["parameterUnits"]
            attrs["long_name"] = msg["parameterName"]
            # add alternative names
            if "cfName" in msg and msg["cfName"] != "unknown":
                attrs["standard_name"] = msg["cfName"]
            if "cfVarName" in msg and msg["cfVarName"] != "unknown":
                attrs["cf_short_name"] = msg["cfVarName"]
            if "shortName" in msg and msg["shortName"] != "unknown":
                attrs["short_name"] = msg["shortName"]

            attrs["grid_type"] = msg["gridType"]
            # information about the rotated pole?
            if msg["gridType"] == "rotated_ll":
                rotated_pole[variable_id] = msg.get_rotated_ll_info(dimension_names[variable_id])
                attrs["grid_mapping"] = rotated_pole[variable_id][0]
            elif msg["gridType"] in ["reduced_gg", "unstructured_grid"]:
                attrs["coordinates"] = "lon lat"

            # grib edition specific attributes
            if msg["editionNumber"] == 1:
                attrs["code"] = msg["indicatorOfParameter"]
                attrs["table"] = msg["table2Version"]
            elif msg["editionNumber"] == 2:
                attrs["discipline"] = msg["discipline"]
                attrs["parameterCategory"] = msg["parameterCategory"]
                attrs["parameterNumber"] = msg["parameterNumber"]

            # store attributes in dictionary for all variables
            attributes[variable_id] = attrs
            # values used for encoding during storage
            encoding = OrderedDict()
            encoding["_FillValue"] = datatype[variable_id](msg["missingValue"])
            encodings[variable_id] = encoding

        # store the values in a dict for later retrieval
        if not in_memory:
            msg_by_var_level_ens[(variable_id, msg["level"], ensemble_member, time_stamp)] = \
                dask.array.from_delayed(dask.delayed(__get_one_message)
                                                    (filename, offset, dimensions[variable_id], datatype[variable_id],
                                                     encodings[variable_id]["_FillValue"]),
                                        shape=dimensions[variable_id],
                                        dtype=datatype[variable_id])
        else:
            # persist the data into memory
            if client is None:
                msg_by_var_level_ens[(variable_id, msg["level"], ensemble_member, time_stamp)] = \
                    dask.array.from_array(msg.get_values(dimensions[variable_id], datatype[variable_id],
                                          encodings[variable_id]["_FillValue"]),
                                          chunks=dimensions[variable_id])
            else:
                if worker is not None:
                    distributed.secede()
                # if all workers are running on local host, load the data directly. There is no advantage of
                # delegating the work to other workers.
                if all_local:
                    chunk = dask.array.from_array(msg.get_values(dimensions[variable_id], datatype[variable_id],
                                              encodings[variable_id]["_FillValue"]),
                                              chunks=dimensions[variable_id])
                    chunk_uploaded = client.scatter(chunk)      # data is currently kept on local worker process.
                    msg_by_var_level_ens[(variable_id, msg["level"], ensemble_member, time_stamp)] = \
                        dask.array.from_delayed(
                                            chunk_uploaded,
                                            shape=dimensions[variable_id],
                                            dtype=datatype[variable_id])
                # we are running with workers distributed of multiple computers.
                # Data reading is done distributed as well.
                else:
                    msg_by_var_level_ens[(variable_id, msg["level"], ensemble_member, time_stamp)] = \
                        client.persist(dask.array.from_delayed(
                                            dask.delayed(__get_one_message)(filename, offset, dimensions[variable_id], datatype[variable_id],
                                                encodings[variable_id]["_FillValue"]),
                                            shape=dimensions[variable_id],
                                            dtype=datatype[variable_id]))
                if worker is not None:
                    distributed.rejoin()

    # close the input file again
    gfile.close()

    logging.debug("finish reading all grib messages from %s, start construction of arrays..." % filename)

    # create the coordinate definition for the dataset
    if len(ensemble_members) > 0:
        coordinates["ens"] = ("ens", numpy.array(sorted(list(ensemble_members)), dtype=numpy.int32))
    coordinates["time"] = ("time", sorted(list(times)))

    # add vertical coordinates
    level_bounds = {}
    level_coordinates = {}
    level_coordinate_names = {}
    for one_var in variables:
        values = level_values[one_var]
        level_coordinate = numpy.zeros((len(values),), dtype=numpy.float32)
        # create bounds of coordinate or only coordinate
        level_bound = None
        if type(values[levels[one_var][0]]) == tuple:
            level_bound = numpy.zeros((len(values), 2), dtype=numpy.float32)
            for ilevel, one_level in enumerate(sorted(levels[one_var])):
                level_bound[ilevel, :] = values[one_level]
                level_coordinate[ilevel] = level_bound[ilevel, :].mean()
        else:
            for ilevel, one_level in enumerate(sorted(levels[one_var])):
                level_coordinate[ilevel] = values[one_level]
        # is the name of the level already in use?
        coordinate_name = one_var[1]
        coordinate_name_counter = 0
        while coordinate_name in level_coordinates:
            if level_coordinates[coordinate_name].size == level_coordinate.size and numpy.all(level_coordinates[coordinate_name] == level_coordinate):
                break
            coordinate_name_counter += 1
            coordinate_name = "%s%d" % (one_var[1], coordinate_name_counter)
        if coordinate_name not in level_coordinates:
            level_coordinates[coordinate_name] = level_coordinate
        level_coordinate_names[one_var] = coordinate_name
        if level_bound is not None:
            level_bounds["%s_bnds" % coordinate_name] = level_bound

    # create the actual data arrays in a loop over all variables
    xarray_variables = OrderedDict()
    for one_var in variables:
        if debug:
            print(one_var)
        # calculate the dimensions for this variable
        var_levels = sorted(levels[one_var])
        var_dims = dimensions[one_var]
        var_shape_ = [len(coordinates["time"][1]), len(ensemble_members), len(var_levels)]
        var_shape_.extend(var_dims)
        var_dim_names_ = ["time", "ens"]
        if one_var in level_coordinate_names:
            var_level_coordinate_name = level_coordinate_names[one_var]
        else:
            var_level_coordinate_name = one_var[1]
        var_dim_names_.append(var_level_coordinate_name)
        var_dim_names_.extend(dimension_names[one_var])
        var_shape = []
        var_dim_names = []
        var_has_meaningful_levels = var_level_coordinate_name in level_coordinates \
                                    and (level_coordinates[var_level_coordinate_name][0] != 0 or
                                         level_coordinates[var_level_coordinate_name].size > 1) \
                                    and ("%dm" % level_coordinates[var_level_coordinate_name][0] not in one_var[0].lower())
        for ishape in range(len(var_shape_)):
            if var_shape_[ishape] > 1 or var_dim_names_[ishape] == "time" or \
                    (var_dim_names_[ishape] in level_coordinates and var_has_meaningful_levels):
                var_shape.append(var_shape_[ishape])
                var_dim_names.append(var_dim_names_[ishape])

        # loop over all timesteps
        stacked_times = []
        for one_time in coordinates["time"][1]:
            # loop over all ensemble members
            stacked_ensemble = []
            if "ens" in coordinates:
                ensemble_members = coordinates["ens"][1]
            else:
                ensemble_members = [-1]
            for one_ens in ensemble_members:
                # loop over all levels
                stacked_levels = []
                for one_level in var_levels:
                    msg_key = (one_var, one_level, one_ens, one_time)
                    if msg_key in msg_by_var_level_ens:
                        stacked_levels.append(msg_by_var_level_ens[(one_var, one_level, one_ens, one_time)])
                    else:
                        stacked_levels.append(dask.array.from_delayed(
                            dask.delayed(numpy.full)(var_dims, encodings[variable_id]["_FillValue"]),
                            shape=dimensions[variable_id],
                            dtype=datatype[variable_id]))

                # combine all levels into one array
                if len(stacked_levels) > 1:
                    stacked_levels = dask.array.stack(stacked_levels)
                elif var_has_meaningful_levels:
                    stacked_levels = dask.array.reshape(stacked_levels[0], (1,) + stacked_levels[0].shape)
                else:
                    stacked_levels = stacked_levels[0]

                # create a list of all ensemble members
                stacked_ensemble.append(stacked_levels)

            # combine all ensemble members into one array
            if len(stacked_ensemble) > 1:
                stacked_ensemble = dask.array.stack(stacked_ensemble)
            else:
                stacked_ensemble = stacked_ensemble[0]

            # create a list of all times
            stacked_times.append(stacked_ensemble)

        # combine all times into one array
        if len(stacked_times) > 1:
            if debug:
                print(stacked_times)
                print(one_var)
            da_var = dask.array.stack(stacked_times)
        else:
            da_var = dask.array.reshape(stacked_times[0], (1,) + stacked_times[0].shape)

        # finally create the xarray and add it to the dataset
        # use the level only in the name if the name is otherwise not unique
        name_is_unique = True
        for one_var_compare in variables:
            if one_var_compare[0] == one_var[0] and one_var_compare[1] != one_var[1]:
                name_is_unique = False
                break
        if name_is_unique:
            var_name = "%s" % one_var[0]
        else:
            var_name = "%s_%s" % (one_var)

        # time, ens, and horizontal coordinates
        var_coords = OrderedDict()
        for coord in coordinates:
            coord_dims = coordinates[coord][0]
            if not type(coord_dims) == list:
                coord_dims = [coord_dims]
            if all(one_dim in var_dim_names for one_dim in coord_dims):
                var_coords[coord] = coordinates[coord]

        # vertical coordinate
        if var_has_meaningful_levels:
            var_coords[var_level_coordinate_name] = level_coordinates[var_level_coordinate_name]

        # create the xarray object
        xarray_variables[var_name] = xarray.DataArray(da_var,
                                                      dims=var_dim_names,
                                                      name=var_name,
                                                      attrs=attributes[one_var],
                                                      coords=var_coords)
        xarray_variables[var_name].encoding = encodings[one_var]

        # are there lon-lat values?
        if "cell" in var_dim_names[-1] and "lon" not in xarray_variables and "lon" in coordinates:
            xarray_variables["lon"] = coordinates["lon"]
            xarray_variables["lat"] = coordinates["lat"]

        # is there a rotated pole for this variable?
        if one_var in rotated_pole:
            if rotated_pole[one_var][0] not in xarray_variables:
                xarray_variables[rotated_pole[one_var][0]] = rotated_pole[one_var][1]
            if var_dim_names[-2] not in xarray_variables:
                xarray_variables[var_dim_names[-2]] = rotated_pole[one_var][2]
            if var_dim_names[-1] not in xarray_variables:
                xarray_variables[var_dim_names[-1]] = rotated_pole[one_var][3]

    # add all bounds arrays to the dataset
    for one_bounds in level_bounds.keys():
        xarray_variables[one_bounds] = xarray.DataArray(level_bounds[one_bounds],
                                                        dims=(one_bounds.replace("_bnds", ""), "bnds"))

    # sort the variables by name
    xarray_variables_sorted = OrderedDict(sorted(xarray_variables.items(), key=lambda i: i[0].lower()))

    # create the dataset object
    logging.debug("creating xarray dataset for %s ..." % filename)
    dataset = xarray.Dataset(xarray_variables_sorted)
    if "lon" in dataset:
        dataset["lon"].attrs["long_name"] = "longitude"
        dataset["lon"].attrs["units"] = "degrees_east"
    if "lat" in dataset:
        dataset["lat"].attrs["long_name"] = "latitude"
        dataset["lat"].attrs["units"] = "degrees_north"

    # data from one ensemble member? store as attribute
    if "ens" in coordinates and len(coordinates["ens"][1]) == 1:
        dataset.attrs["ensemble_member"] = coordinates["ens"][1][0]

    # add bounds attribute to coordinates
    for one_bounds in level_bounds.keys():
        coordinate_name = one_bounds.replace("_bnds", "")
        if coordinate_name in dataset:
            dataset[coordinate_name].attrs["bounds"] = one_bounds
    logging.debug("finished reading %s" % filename)

    return dataset


def __get_one_message(filename, offset, shape, dtype, missing):
    """
    get the values of the message at position *imsg* from a grib file

    Parameters
    ----------
    filename
    imsg

    Returns
    -------
    numpy.ndarray
    """
    # open the input file, seek the message and read it
    with io.open(filename, "rb") as gfile:
        # read and decode the message
        msg = eccodes_cffi.GribMessage(gfile, offset, read_data=True)

        # decode the actual values
        return msg.get_values(shape, dtype, missing)



import logging
import os
from pathlib import Path
from datetime import datetime
from urllib.error import HTTPError
from urllib.request import urlopen
import pandas
from enstools.misc import download, bytes2human
from enstools.core import get_cache_dir


class DWDRadar:
    # create a new temporal directory to store content.log and opendata_dwd_content.pkl
    cache_path = get_cache_dir()

    def __init__(self, refresh_content=False):

        def create_dataframe(logdata):
            """
            Creates a pandas.DataFrame from the given log data from https://opendata.dwd.de/weather/radar/
            and sets the following attributes:
                file: str
                    Name of the url of the file.
                size: str
                    The size of the file.
                time: np.datetime64
                    The creation time on the server.
                product_class: str
                    The radar product_class, e.g. "radolan".
                format: str
                    The format of the file.
                filename: str
                    The name of the file.
                forecast_time: int
                    Minutes since the initalization of the forecast.
                    If the radar product is not a forecast this is alwys zero.

            Parameters
            ----------
            logdata: str
                The name of the content logfile.

            Returns
            -------
            content: pandas.DataFrame
                The DataFrame object with the given attributes.

            """
            logging.info("Creating content database with {}".format(logdata))
            content = pandas.read_csv(logdata, delimiter="|", header=None, names=["file", "size", "upload_time"])
            content["upload_time"] = pandas.to_datetime(content["upload_time"], format="%Y-%m-%d %H:%M:%S")
            content["size"] = content["size"].astype(int)

            # ignore the content.log file and tar.gz files
            content = content[~content.file.str.contains("content.log")]
            content = content[~content.file.str.contains(".tar.")]

            content["product_class"] = content["file"].apply(lambda x: x.split("/")[1])
            content["product"] = content["file"].apply(lambda x: x.split("/")[2])
            content = content[(content.product_class == "composit") |
                              (content.product_class == "radvor") |
                              (content.product_class == "radolan")]
            content["filename"] = content["file"].apply(lambda x: x.split("/")[3])
            content["data_time"] = content["filename"]
            content["data_time"] = content["data_time"].apply(lambda x: datetime.strptime(x[15:-10], "%y%m%d%H%M")
                                                              if x.endswith("-dwd---bin")
                                                              else datetime.strptime(x[2:-7], "%y%m%d%H%M")
                                                              if (x.startswith("RE") | x.startswith("RQ"))
                                                              else datetime.strptime(x[2:-8], "%y%m%d%H%M")
                                                              if (x.startswith("FX") | x.startswith("WN"))
                                                              else self.__scrape_pg_paah_time(x)
                                                              if x.startswith("PAAH")
                                                              else datetime.strptime(x[15:-11], "%y%m%d%H%M")
                                                              if x.endswith("-dwd---bufr")
                                                              else None)
            content["forecast_time"] = content["filename"]
            content["forecast_time"] = content["forecast_time"].apply(lambda x: int(x[-6:-3])
                                                                      if (x.startswith("RE")
                                                                          or x.startswith("RQ"))
                                                                      else int(x[-7:-4])
                                                                      if x.startswith("WN")
                                                                      else 0)
            content["format"] = content["filename"]
            content["format"] = content["format"].apply(lambda x:
                                                        "bin" if x.endswith("bin")
                                                        else "bufr" if x.endswith("bufr")
                                                        else "gz" if x.endswith("gz")
                                                        else "buf" if x.endswith(".buf")
                                                        else ".bz2" if x.endswith(".bz2")
                                                        else None)

            content["file"] = content["file"].apply(lambda x: "https://opendata.dwd.de/weather/radar" + x[1:])

            content.to_pickle(self.content_pkl_path)
            return content

        self.content_pkl_path = os.path.join(DWDRadar.cache_path, "opendata_dwd_content_radar.pkl")
        if os.path.exists(self.content_pkl_path) and not refresh_content:
            logging.info(f"Reading content database from {self.content_pkl_path}")
            content_old = pandas.read_pickle(self.content_pkl_path)
            self.content = content_old

        else:
            if refresh_content:
                logging.info("Refreshing content database")
            else:
                logging.info("Initializing content database")
            if os.path.exists(os.path.join(DWDRadar.cache_path, "content.log")):
                os.remove(os.path.join(DWDRadar.cache_path, "content.log"))
            self.content_log_path = os.path.join(DWDRadar.cache_path, "content_radar.log.bz2")
            download("https://opendata.dwd.de/weather/radar/content.log.bz2",
                     destination=self.content_log_path, uncompress=False)
            self.content = create_dataframe(self.content_log_path)

    @staticmethod
    def __scrape_pg_paah_time(filename):
        """
        Determine the missing year and month in PAAH filenames from todays date.
        """
        date_now = datetime.utcnow()
        file_date = datetime.strptime(filename[10:-4], "%d%H%M")

        # If somebody wants to download data after new year party:
        if date_now.month == 1 and date_now.day <= 3:
            if file_date.day >= 28:
                file_date = file_date.replace(year=date_now.year-1, month=12)
        else:
            file_date = file_date.replace(year=date_now.year, month=date_now.month)
        return file_date

    def refresh_content(self):
        """
        Initializes the DWDRadar object again.
        Downloads the actual content.log from the server and creates a new DataFrame.
        """
        self.__init__(refresh_content=True)

    def get_product_classes(self):
        """
        Gives the available product classes of the data server.
        Returns
        -------
            List of Strings with the available product classes

        """
        content = self.content
        avail_product_classes = content["product_class"].drop_duplicates().values.tolist()
        avail_product_classes.sort()
        return avail_product_classes

    def get_products(self, product_class=None):
        """
        Gives the available radar products.
        Parameters
        ----------
        product_class: string
            The group of the radar data products.

        Returns
        -------
        If available, the right data products will be returned.
        """
        content = self.content
        if product_class is None:
            avail_products = content["product"].drop_duplicates().values.tolist()
        else:
            avail_products = content[content["product_class"] == product_class]["product"]\
                .drop_duplicates().values.tolist()
        return avail_products

    def get_avail_data_times(self, product=None):
        """
        Gives the available data times for a given radar product.

        Parameters
        ----------
        product: str
            The name of the data product.

        Returns
        -------
        avail_data_times: list
            A list of integers of the available data times.

        """
        content = self.content
        avail_data_times = content[(content["product"] == product)]["data_time"].drop_duplicates().dt\
            .to_pydatetime().tolist()
        avail_data_times.sort()
        return avail_data_times

    def get_avail_forecast_times(self, product=None, data_time=None):
        """
        Gives the available  forecast times since the data_time for a given radar product and data_time.

        Parameters
        ----------
        product: str
            The name of the data product.
        data_time: int
            The time of the radar data for which the available forecast times want to be known.

        Returns
        -------
        avail_forecast_times: list
            The available times of the radar data since the initialization of the forecast.

        """
        content = self.content
        avail_forecast_times = content[(content["product"] == product)
                                       & (content["data_time"] == data_time)]["forecast_time"] \
            .drop_duplicates().values.tolist()
        avail_forecast_times.sort()
        return avail_forecast_times

    def get_avail_file_formats(self, product=None):
        """
        Gives the available file formats for a given radar product.
        Currently most products only have one format to choose.

        Parameters
        ----------
        product: str
            The name of the data product.

        Returns
        -------
        A list of strings of the available formats.

        """
        content = self.content
        avail_file_formats = content[content["product"] == product]["format"].drop_duplicates().values.tolist()
        avail_file_formats.sort()
        return avail_file_formats

    def get_url(self, product=None, data_time=None, forecast_time=None, file_format=None):
        """
        Gives the url of the file on the https://opendata.dwd.de/weather/radar server.

        Parameters
        ----------
        product: str
        data_time: int
            The time of the radar data file for which the url wants to be known.
        forecast_time: int
            The minutes since the data_time of the radar data of the file for which the url wants to be known.
        file_format: str
            The file format (eg. 'gz').

        Returns
        -------
        url: str
            The url adress of the file.
        """
        content = self.content

        url = content[(content["product"] == product)
                      & (content["data_time"] == data_time)
                      & (content["forecast_time"] == forecast_time)
                      & (content["format"] == file_format)]["file"].values
        if len(url) != 1:
            raise KeyError(f"Url not found(product:{product}, data_time:{data_time}, "
                           f"forecast_time:{forecast_time}, format:{file_format})")

        return url.item()

    def get_filename(self, product=None, data_time=None, forecast_time=None, file_format=None):
        """
         Gives the filename of the file on the https://opendata.dwd.de/weather/radar server.

        Parameters
        ----------
        product: str
            The name of the data product.
        data_time: datetime.datetime
            The time of the radar data file for which the filename wants to be known.
        forecast_time: int
            The minutes since the data_time of the radar data of the file for which the filename wants to be known.
        file_format: str
            The file format (eg. '.gz')
        Returns
        -------
        url: str
            The url adress of the file.
        """
        content = self.content

        filename = content[(content["product"] == product)
                           & (content["data_time"] == data_time)
                           & (content["forecast_time"] == forecast_time)
                           & (content["format"] == file_format)]["filename"].values.item()

        return filename

    def check_url_available(self, product=None, data_time=None, forecast_time=None, file_format=None):
        try:
            http_code = urlopen(self.get_url(product=product, data_time=data_time,
                                             forecast_time=forecast_time, file_format=file_format)).getcode()
            if http_code == 200:
                url_available = True
            else:
                url_available = False
        except HTTPError:
            url_available = False
        return url_available

    def get_file_size(self, product=None, data_time=None, forecast_time=None, file_format=None):
        content = self.content
        file_size = content[(content["product"] == product)
                            & (content["data_time"] == data_time)
                            & (content["forecast_time"] == forecast_time)
                            & (content["format"] == file_format)]["size"].values.item()
        return file_size

    def get_size_of_download(self, product=None, data_time=None, forecast_time=None, file_format=None):
        total_size = 0
        for dtime in data_time:
            for ftime in forecast_time:
                total_size += self.get_file_size(product=product, data_time=dtime,
                                                 forecast_time=ftime, file_format=file_format)
        return total_size

    def check_parameters(self, product=None, data_time=None, forecast_time=None, file_format=None):
        """
        Checks if there are all files available for the given parameters.
        If not, the DWDRadar object will be refreshed.
        If one file is not available for the given parameters, a detailed error will be thrown.

        Parameters
        ----------
        product: str
            The name of the data product.
        data_time: list
            The time of the radar data file.
        forecast_time: list
            The minutes since the data_time of the radar data.
        file_format: str
            The file format (eg. 'gz').

        Returns
        -------
        None
        """

        params_available = True

        if product not in self.get_products():
            params_available = False

        for dtime in data_time:
            avail_dtimes = self.get_avail_data_times(product=product)
            if dtime not in avail_dtimes:
                params_available = False
                break

            for ftime in forecast_time:
                if ftime not in self.get_avail_forecast_times(product=product, data_time=dtime):
                    params_available = False
                    break

        if not params_available:
            logging.warning("Parameters not available")
        else:
            for dtime in data_time:
                for ftime in forecast_time:
                    if not self.check_url_available(product=product, data_time=dtime,
                                                    forecast_time=ftime, file_format=file_format):
                        params_available = False
                        break
            if not params_available:
                logging.warning("Database outdated.")

        if not params_available:
            self.refresh_content()

            for dtime in data_time:
                avail_dtimes = self.get_avail_data_times(product=product)
                if dtime not in avail_dtimes:
                    raise ValueError(
                        "The data time {} is not available for the product {}. "
                        "Available variables:{}".format(dtime, product, avail_dtimes))

                avail_forecast_times = self.get_avail_forecast_times(product=product, data_time=dtime)
                for ftime in avail_forecast_times:

                    if ftime not in avail_forecast_times:
                        raise ValueError("The forecast time {} is not available for the data time {}. "
                                         "Possible values:{}".format(ftime, dtime, avail_forecast_times))

    def retrieve(self, product=None, data_time=None, forecast_time=0, dest=None, file_format=None):
        """
        Downloads datasets from opendata server.
        Parameters
        ----------
        product: str
            The type of the radar data.

        data_time : list or int
            Time of the radar data. Multiple values are allowed.

        forecast_time : list or int
                time since the initialization of forecast. Multiple values are allowed.

        file_format : str
            The file format (eg. 'gz'). Only has to be specified when there are more than one available.

        dest : str
                Destination folder for downloaded data. If the files are already available,
                they are not downloaded again.

        Returns
        -------
        list :
                names of downloaded files.
        """
        # Want to download data for one or more times?
        if not isinstance(data_time, (list, tuple)):
            data_time = [data_time]
        # Want to download one or more forecast times?
        if not isinstance(forecast_time, (list, tuple)):
            forecast_time = [forecast_time]

        if not os.path.exists(dest):
            os.mkdir(dest)
        if product is not None:
            product = product.lower()

        if file_format is None:
            avail_formats = self.get_avail_file_formats(product=product)
            if len(avail_formats) == 1:
                file_format = avail_formats[0]
            else:
                raise KeyError(f"The format for the product {product} is ambigous."
                               " You have to specify it in the 'format' keyword")

        download_files = []
        download_urls = []

        self.check_parameters(product=product, data_time=data_time,
                              forecast_time=forecast_time, file_format=file_format)
        for dtime in data_time:
            for ftime in forecast_time:
                download_urls.append(self.get_url(product=product, data_time=dtime,
                                                  forecast_time=ftime, file_format=file_format))
                download_files.append(self.get_filename(product=product, data_time=dtime,
                                                        forecast_time=ftime, file_format=file_format))

        download_files = [os.path.join(dest, file) for file in download_files]

        total_size_human = bytes2human(self.get_size_of_download(product=product,
                                                                 data_time=data_time,
                                                                 forecast_time=forecast_time,
                                                                 file_format=file_format))
        logging.info("Downloading {} files with the total size of {}".format(len(download_files), total_size_human))
        for i in range(len(download_urls)):
            download(download_urls[i], download_files[i], uncompress=False)

        return download_files


# store the created content object for later re-use
__content = None


def getDWDRadar():
    """
    Creates a DWDRadar object and returns it. Multiple calls to this function will return the same object.

    Returns
    -------
    DWDRadar
    """
    global __content
    if __content is not None:
        return __content
    else:
        __content = DWDRadar()
        return __content



"""
Reading and Writing of meteorological data
"""


def __clean_HDF5_PLUGIN_PATH():
    """
    if the libraries from hdf5plugin are in HDF5_PLUGIN_PATH, then remove them
    """
    import os
    import logging
    if "HDF5_PLUGIN_PATH" in os.environ:
        paths = os.environ["HDF5_PLUGIN_PATH"].split(":")
        keep = []
        for one_path in paths:
            if len(one_path) == 0:
                continue
            if 'h5z-sz' not in one_path:
                logging.info(f"removed {one_path} from HDF5_PLUGIN_PATH")
                continue
            keep.append(one_path)
        if len(keep) > 0:
            os.environ["HDF5_PLUGIN_PATH"] = ":".join(keep)
        else:
            del os.environ["HDF5_PLUGIN_PATH"]


# TODO: figure out why this is needed and remove it!
# __clean_HDF5_PLUGIN_PATH()

from .file_type import get_file_type
from .reader import read
from .writer import write
from .dataset import drop_unused




import numpy as np
from scipy import ndimage
from enstools.core import check_arguments
from enstools.interpolation import downsize


@check_arguments(shape={"im": (0, 0)})
def embed_image(im, fac=None, sRow=None, sCol=None):
    """
    embed an array into one that has multiples of fac as sizes; fill missing rows/columns with zeros

    Parameters
    ----------
    im : xarray.DataArray or np.ndarray
            data array, 2d

    fac : int
            factor - new array has multiples of fac as sizes in both dimensions

    sRow : int
            defines # of rows im has to be divisible by

    sCol : int
            defines # of columns im has to be divisible by

    Returns
    -------
    xarray.DataArray
            array with multiples of fac in size, im embedded
    """
    nRows, nColumns = im.shape
    m, n = nRows, nColumns
    # define next bigger size with m,n % fac == 0
    if fac is not None:
        m, n = ((np.array([m, n]) - 1) // int(fac) + 1) * int(fac)
    if sRow is not None:
        m = (m // int(sRow) + 1) * int(m)
    if sCol is not None:
        n = (n // int(sCol) + 1) * int(sCol)
    if m > nRows or n > nColumns:
        im_handle = np.zeros((m, n), dtype=im.dtype)
        im_handle[:nRows, :nColumns] = im
        return im_handle
    return im


@check_arguments(shape={"image": (0, 0), "xdis": "image", "ydis": "image"})
def map_backward(image, xdis, ydis):
    """
    An image is created by selecting for each position the values of the
    input image at the positions given by the xdis(placement) and ydis(placement).
    The routine works backward - the displacement vectors have to refer to their destination
    - as realized in match_meteosat.pro Hermann Mannstein, May 2003

    Parameters
    ----------
    image : xarray.DataArray or np.ndarray
            image to morph

    xdis : xarray.DataArray or np.ndarray
            displacement in x-direction for each point

    ydis : xarray.DataArray or np.ndarray
            displacement in y-direction for each point

    Returns
    -------
    xarray.DataArray or np.ndarray
            morphed image
    """
    nx = np.shape(image)[1]
    n = np.size(image)

    # seperate displacement in integer (lower value:floor() )  and float values
    # notice: displacement vectors can point in between an area of four surrounding elements
    ixdis = np.array(np.floor(xdis), dtype=int)
    iydis = np.array(np.floor(ydis), dtype=int)
    fxdis = xdis - ixdis
    fydis = ydis - iydis
    fxydis = fxdis * fydis

    # define indexarray for pixel displacement
    ind = np.resize(np.arange(n), np.shape(ixdis))
    # add displacement values to indexarray
    ind = (ind + n + ixdis + iydis * nx) % n  # modulo n for borders of image: wrap around edges

    # now displace pixel image.take(ind)
    # therefor take four surrounding values and weight them with the length of the displacement vectors
    # (weighting factor is float value of displacement vector)
    imdis = (
        (1. - fxdis - fydis + fxydis) * image.take(
            ind) +  # value where the floor value of displacement vector points at
        (fxdis - fxydis) * image.take((ind + n + 1) % n) +  # value right of floor value
        (fydis - fxydis) * image.take((ind + n + nx) % n) +  # value below floor value
        fxydis * image.take((ind + n + nx + 1) % n))  # value right below floor value
    return imdis


def gauss_kern(size, sigma):
    """
    Returns a normalized 2D gauss kernel array for convolutions

    Parameters
    ----------
    size : int
            size of kernel (size=2 refers to a 5x5 kernel)

    sigma : float
            standard deviation of gauss kernel

    Returns
    -------
    array for convolution
    """
    size = int(size)
    x, y = np.mgrid[-size:size + 1, -size:size + 1]
    g = np.exp(-(x ** 2 / float(size) + y ** 2 / float(size)) / sigma ** 2)
    return g / g.sum()


@check_arguments(shape={"image1": (0, 0), "image2": (0, 0)})
def match_pyramid(image1, image2, factor=4, sigma=5 / 3.):
    """
    Morph image1 that it best matches image2 with pyramidal image matching.

    Parameters
    ----------
    image1 : xarray.DataArray or np.ndarray
            image to morph

    image2 : xarray.DataArray or np.ndarray
            image to match

    factor :  int
            subsampling factor or topmost pyramid level with lowest resolution (averaging 2^factor data points onto one) (default: 4)

    sigma : float
            standard deviation of gauss kernel (default: 5/3.)

    Returns
    -------
    tuple
            morphed image, x-displacement and y-displacement of each data point and least square error of morphed image.
    """
    image1 = np.array(image1, dtype=image1.dtype)
    image2 = np.array(image2, dtype=image2.dtype)
    oS = np.shape(image1)

    # make sure images have powers of 2 as sizes
    im1 = embed_image(image1, 2 ** factor)
    im2 = embed_image(image2, 2 ** factor)
    f1 = im1.copy()

    # initialize displacement vectors
    xdis = np.zeros(np.shape(im1), dtype=np.float32)
    ydis = np.zeros(np.shape(im1), dtype=np.float32)

    ke = gauss_kern(2, sigma)
    zoom = ndimage.zoom  # resize and array with cubic spline interpolation with given order
    convolve = ndimage.convolve  # convolve and array with given weights

    # start pyramide
    for s in range(factor, -1, -1):
        zoomfac = 2 ** s
        n1 = downsize(f1, zoomfac)
        n2 = downsize(im2, zoomfac)

        xd = np.zeros(np.shape(n1), dtype=int)
        yd = np.zeros(np.shape(n1), dtype=int)
        qu = convolve(((n1 - n2) ** 2).astype(n1.dtype), ke, mode='nearest')

        # search in area (-2 to 2) for best dispacement
        for i in range(-2, 3):
            for j in range(-2, 3):
                n1s = np.roll(np.roll(n1, i, axis=1), j, axis=0)  # shift image array, wrap ends around edges
                sm = convolve(((n1s - n2) ** 2).astype(n1s.dtype), ke, mode='nearest')
                ind = np.where((sm < qu))
                xd[ind] = i
                yd[ind] = j
                qu[ind] = sm[ind]

        xdis -= zoom(convolve(xd.astype(np.float32) * zoomfac, ke, mode='constant'), zoomfac,
                     order=1)  # resize using linear interpolation
        ydis -= zoom(convolve(yd.astype(np.float32) * zoomfac, ke, mode='constant'), zoomfac, order=1)

        f1 = map_backward(im1, xdis, ydis)

        lse = convolve(qu, ke, mode='nearest')[:oS[0], :oS[1]]
    #	lse = convolve(((f1-im2)**2).astype(np.float), gauss_kern(2,.5), mode='nearest')[:oS[0],:oS[1]]
    #	lse = ((f1-im2)**2).astype(np.float)[:oS[0],:oS[1]]
    f1 = f1[:oS[0], :oS[1]]
    xdis = -xdis[:oS[0], :oS[1]]
    ydis = -ydis[:oS[0], :oS[1]]

    return f1, xdis, ydis, lse



"""
helper functions for automatic parallelization using dask
"""
from decorator import decorate, decorator
from enstools.misc import get_time_dim, get_ensemble_dim
import xarray
import dask
from distributed import wait
import numpy as np
import logging
import six
from enstools.core import get_arg_spec
import inspect


def __args_to_dask(*args):
    """
    convert arguments to dask arrays or extract dask arrays if already present.

    Parameters
    ----------
    args :
            numpy, xarray, or dask array

    Returns
    -------
    args :
            all arguments converted to dask
    """
    new_args = []
    for one_arg in args:
        # how should be re-chunked?
        if isinstance(one_arg, dask.array.core.Array):
            new_args.append(one_arg)
        elif isinstance(one_arg, xarray.core.dataset.Dataset):
            # it is a dataset, check all data variables individually
            # first pass: are there any none-dask arrays?
            has_none_dask = False
            for one_data_var_name, one_data_var in six.iteritems(one_arg.data_vars):
                if not isinstance(one_data_var, dask.array.core.Array):
                    has_none_dask = True
                    break
            # we have none-dask arrays, create a copy one the dataset and replace single variables
            if has_none_dask:
                new_one_arg = one_arg.copy()
                for one_data_var_name, one_data_var in six.iteritems(one_arg.data_vars):
                    if not isinstance(one_data_var.data, dask.array.core.Array):
                        new_one_data_var = __args_to_dask(one_data_var)[0]
                        new_one_arg[one_data_var_name].data = new_one_data_var
                new_args.append(new_one_arg)
            else:
                new_args.append(one_arg)
        elif isinstance(one_arg, xarray.core.dataarray.DataArray):
            if isinstance(one_arg.data, dask.array.core.Array):
                # if the data is already a dask array, wait for it to become ready
                try:
                    wait(one_arg.data)
                except ValueError:
                    logging.debug("not client found, can not wait for computations on the cluster!")
                new_args.append(one_arg.data)
            else:
                new_args.append(one_arg.chunk().data)
        elif isinstance(one_arg, np.ndarray):
            new_args.append(dask.array.from_array(one_arg, chunks=-1))
        else:
            new_args.append(one_arg)
    return new_args


def rechunk_arguments(dim_type=None, dim_name=None, arg_names=None):
    """
    rechunk all arguments along a specific dimensions

    Parameters
    ----------
    arg : list
            list of arguments. all of them must be xarray DataArrays or Datasets

    dim_type : dict
            keys: dimensions types
            "time": all known names for time dimensions are used
            "ens": all known names for ensemble dimensions are used
            values: chunk size along this dimension

    dim_name : dict
            use exactly this dimension name, do not search for dimensions of this type with other names.

    arg_names : list
            list of arguments to process. None: all arguments without defaults. This argument is not yet implemented!
    """
    # check decorator arguments
    if dim_type is None and dim_name is None:
        raise ValueError("the decorator rechunk_arguments needs at least one of the arguments dim_type or dim_name!")
    if dim_type is None:
        dim_type = {}
    if dim_name is None:
        dim_name = {}

    @decorator
    def function_wrapper(func, *args, **kwargs):
        # check the actual values of the functions arguments
        arg_spec = get_arg_spec(func)
        arg_names = arg_spec[0]
        defaults = arg_spec[3]
        arg_values = inspect.getcallargs(func, *args, **kwargs)

        # create a list of the new arguments
        new_args = []

        # loop over all none-keyword arguments
        for iarg, one_arg_name in enumerate(arg_names):
            one_arg = arg_values[one_arg_name]
            if iarg >= len(arg_names) - len(defaults):
                pass
            elif isinstance(one_arg, xarray.core.dataarray.DataArray) or isinstance(one_arg, xarray.core.dataset.Dataset):
                chunks = {}
                # find the dimension to re-chunk
                if "time" in dim_type:
                    dn = get_time_dim(one_arg)
                    chunks[dn] = dim_type["time"]
                if "ens" in dim_type:
                    dn = get_ensemble_dim(one_arg)
                    chunks[dn] = dim_type["ens"]
                # loop over all explicitly named dimensions
                for dn, cs in six.iteritems(dim_name):
                    if dn in one_arg:
                        chunks[dn] = cs
                # apply the re-chunk operation
                one_arg = one_arg.chunk(chunks)
            # allow scalar arguments
            elif isinstance(one_arg, int) or isinstance(one_arg, float):
                pass
            else:
                raise ValueError("automatic re-chunking is only possible with xarray arguments!")

            # construct new argument list
            if iarg > len(new_args) and one_arg_name in kwargs:
                kwargs[one_arg_name] = one_arg
            else:
                new_args.append(one_arg)
                if one_arg_name in kwargs:
                    del kwargs[one_arg_name]

        return func(*new_args, **kwargs)

    return function_wrapper


def apply_chunkwise(func):
    """
    Automatic parallelisation of a decorated function. All arguments have to have the same shape!
    """

    def function_wrapper(func, *args, **kwargs):
        dask_args = __args_to_dask(*args)
        result = dask.array.map_blocks(func, *dask_args, dtype=args[0].dtype)
        return result

    return decorate(func, function_wrapper)



import xarray as xr
from collections import OrderedDict
from six import iteritems


def __swap_dims_n(self, dims_dict, inplace=False):
    for k, v in dims_dict.items():
        if k not in self.dims:
            raise ValueError('cannot swap from dimension %r because it is '
                             'not an existing dimension' % k)

    result_dims = set(dims_dict.get(dim, dim) for dim in self.dims)

    variables = OrderedDict()

    coord_names = self._coord_names.copy()
    coord_names.update(dims_dict.values())

    for k, v in iteritems(self.variables):
        dims = tuple(dims_dict.get(dim, dim) for dim in v.dims)
        if k in result_dims:
            var = v.to_index_variable()
        else:
            var = v.to_base_variable()
        var.dims = dims
        variables[k] = var

    return self._replace_vars_and_dims(variables, coord_names, inplace=inplace)


def __staggered_grid_interpolation(data, src_coord, dst_coord, label='upper'):
    """
    perform the actual interpolation on a Dataset
    
    Parameters
    ----------
    data : xarray.Dataset
            dataset with staggered variables

    src_coord : str
            name of staggered coordinate to unstagger

    dst_coord : str
            name of unstaggered destination variable

    label : {"upper", "lower"}
            upper: first value along the unstaggered dimension will be missing, all upper values will be present.
            lower: last value along the unstaggered dimension will be missing, all lower values will be present.

    Returns
    -------
    xarray.Dataset :
            unstaggered Dataset.
    """
    if isinstance(data, xr.core.dataset.Dataset):
        variables = OrderedDict()
        # prepare slices
        kwargs_start = {src_coord: slice(None, -1)}
        kwargs_end = {src_coord: slice(1, None)}
        # prepare new coordinate
        if label == 'upper':
            kwargs_new = kwargs_end  # {dst_coord: slice(1, None)}
        elif label == 'lower':
            kwargs_new = kwargs_start  # {dst_coord: slice(None, -1)}
        else:
            raise ValueError('The \'label\' argument has to be either '
                             '\'upper\' or \'lower\'')

        if dst_coord not in data.variables:
            variables[dst_coord] = ((data.variables[src_coord].isel(**kwargs_end) + data.variables[src_coord].isel(**kwargs_start)) / 2.)
            variables[dst_coord].dims = (dst_coord,)
        for name, var in iteritems(data.variables):
            if src_coord in var.dims:
                if name in data.data_vars:
                    variables[name] = ((var.isel(**kwargs_end) + var.isel(**kwargs_start)) / 2.)
                else:
                    variables[name] = var.isel(**kwargs_new)
            else:
                variables[name] = var
        result = data._replace_vars_and_dims(variables)
        return result
    else:
        raise TypeError('FileType for __staggered_grid_interpolation has to be either a xarray Dataset')


def unstagger(data):
    """
    Interpolate variables on staggered grid (e.g., U and V) to grid cell centers. Every model has different names for
    staggered coordinates and variables. Currently, only the names of the model COSMO are known and supported.

    Parameters
    ----------
    data : xarray.DataArray or xarray.DataSet
            Array or Dataset with staggered variables. For datasets, all staggered variables are unstaggered. The
            staggered coordinate arrays are removed

    Returns
    -------
    xarray.DataArray or xarray.DataSet
            The return type depends on the input type.
    """
    # create a copy of the input to ensure that the input remains unchanged
    data = data.copy(deep=False)

    # check variable and select variable names
    staggered_coords = {}
    if "srlon" in data.coords:
        staggered_coords["srlon"] = ("rlon", "upper", "slonu", "slatu")
    if "srlat" in data.coords:
        staggered_coords["srlat"] = ("rlat", "upper", "slonv", "slatv")
    if len(staggered_coords) == 0:
        raise ValueError("no supported staggered coordinates found in input of function unstagger!")

    # loop over all staggered coordinates.
    for staggered_coord, unstaggered_coord in iteritems(staggered_coords):
        if isinstance(data, xr.core.dataset.Dataset):
            for name, var in iteritems(data.data_vars):
                if staggered_coord in var.dims:
                    ds = var._to_temp_dataset()
                    ds[unstaggered_coord[0]] = data[unstaggered_coord[0]]
                    ds = __staggered_grid_interpolation(ds, staggered_coord, unstaggered_coord[0], label=unstaggered_coord[1])
                    data[name] = var._from_temp_dataset(ds)
            data = __swap_dims_n(data, {staggered_coord: unstaggered_coord[0]})
        elif isinstance(data, xr.core.dataarray.DataArray):
            ds = __staggered_grid_interpolation(data._to_temp_dataset(), staggered_coord, unstaggered_coord[0], label=unstaggered_coord[1])
            ds = __swap_dims_n(ds, {staggered_coord: unstaggered_coord[0]})
            data = data._from_temp_dataset(ds)
        else:
            raise TypeError('Input type for unstagger has to be either a xarray.Dataset or xarray.DataArray')

    # remove staggered coordinates
    for staggered_coord, unstaggered_coord in iteritems(staggered_coords):
        for to_remove in unstaggered_coord[2:]:
            if to_remove in data.coords:
                del data[to_remove]
        if staggered_coord in data.coords:
            del data[staggered_coord]

    return data



from numba import jit
import numpy as np
import xarray
from enstools.misc import swapaxis
from collections import OrderedDict
import logging


class model2pressure:

    def __init__(self, src_p, dst_p, vertical_dim=None):
        """
        Create an interpolator object for the interpolation from model to pressure level

        Parameters
        ----------
        src_p : xarray.DataArray or np.ndarray
                full model pressure at all grid cells. 2d for unstructured grid and 3d for regular grid. The rightmost
                dimension is expected to be the level coordinate, e.g. (lev, lat, lon) or (lev, cell)

        dst_p : xarray.DataArray or np.ndarray or float
                pressure level(s) to interpolate to

        vertical_dim : int
                the position of the vertical axes in the input data. if not specified, it is automatically detected. In
                this case, for numpy arrays, it has to be the first axes.

        Returns
        -------
        model2pressure
                callable interpolator object.
        """

        # find the vertical dimension of the input array
        if vertical_dim is None:
            if isinstance(src_p, xarray.DataArray):
                vertical_dim_names = ["pres", "p", "lev", "level", "isobaric", "layer", "hybrid", "height", "height_2"]
                for dimi, dim in enumerate(src_p.dims):
                    for valid_name in vertical_dim_names:
                        if valid_name in dim.lower():
                            vertical_dim = dimi
                            break
                    if vertical_dim is not None:
                        break
                if vertical_dim is not None:
                    logging.info("using dimension %d (%s) as vertical dimension." % (vertical_dim, src_p.dims[vertical_dim]))
                else:
                    raise ValueError("unable to auto-detect vertical dimension in input data.")
            else:
                if src_p.ndim > 3:
                    raise ValueError("unsupported number of dimensions for numpy array: %d. Try (lev, lat, lon) or (lev, cell) or xarray." % src_p.ndim)
                else:
                    vertical_dim = 0
        self._src_vertical_dim = vertical_dim

        # reorder the array if the vertical dimension is not the first dimension
        self._src_shape_not_reordered = src_p.shape
        if self._src_vertical_dim > 0:
            src_p = swapaxis(src_p, self._src_vertical_dim, 0)

        # is the destination a single value?
        if not hasattr(dst_p, "__len__"):
            dst_p = np.asarray([dst_p], dtype=np.float64)

        # store the original shape of the source field and the expected destination field
        self._src_shape = src_p.shape
        self._dst_shape = (len(dst_p),) + src_p.shape[1:]

        # store attributes and coordinates of the source array
        if isinstance(src_p, xarray.DataArray):
            self._src_attrs = src_p.attrs
            self._src_dims = src_p.dims
            self._src_coords = src_p.coords
        else:
            self._src_attrs = OrderedDict()
            self._src_dims = ("lev",)
            for dim in range(1, src_p.ndim):
                self._src_dims += ("dim_%d" % dim,)
            self._src_coords = OrderedDict()

        # store the destination pressure as coordiinate for the result arrays
        self._dst_pressure = dst_p

        # if the input pressure field is 3d, then flatten to two leftmost dimensions
        if src_p.ndim > 2:
            src_p = np.asarray(src_p).reshape((src_p.shape[0], np.product(src_p.shape[1:])))

        # calculate the weights for each horizontal grid point
        self._indices, self._weights = get_weights(np.asarray(src_p), np.asarray(dst_p))

    def __call__(self, data):
        """
        perform the actual interpolation

        Parameters
        ----------
        data : xarray.DataArray or np.ndarray
                the data array to interpolate. The shape has to match the shape of the array used to create this object

        Returns
        -------
        xarray.DataArray
                the interpolated array (lev, lat, lon) or (lev, cell)
        """
        # check the shape of the input array
        if data.shape != self._src_shape_not_reordered:
            raise ValueError("the shape of the data array to interpolate has to match the shape of the original pressure array: %s" % str(self._src_shape_not_reordered))

        # are there attributes at the data field?
        if isinstance(data, xarray.DataArray):
            data_attrs = data.attrs
            data_name = data.name
        else:
            data_attrs = OrderedDict()
            data_name = "interpolated"

        # reorder the dimensions if required
        if self._src_vertical_dim > 0:
            data = swapaxis(data, self._src_vertical_dim, 0)

        # if the input data field is 3d, then flatten to two leftmost dimensions
        if data.ndim > 2:
            data = np.asarray(data).reshape((data.shape[0], np.product(data.shape[1:])))

        # perform the interpolation
        result_data = apply_weights(np.asarray(data), self._indices, self._weights)

        # reshape required?
        if result_data.shape != self._dst_shape:
            result_data = result_data.reshape(self._dst_shape)

        # create a xarray object
        result = xarray.DataArray(result_data, dims=("pressure",) + self._src_dims[1:], attrs=data_attrs, name=data_name)
        result.coords["pressure"] = self._dst_pressure
        # copy coordinates for the rightmost dimensions
        for dim in self._src_dims[1:]:
            if dim in self._src_coords:
                result.coords[dim] = self._src_coords[dim]

        # reorder required?
        if self._src_vertical_dim > 0:
            result = swapaxis(result, self._src_vertical_dim, 0)
        return result


@jit(nopython=True)
def apply_weights(data, indices, weights):
    """
    use the indices and weights to create a new interpolated array

    Parameters
    ----------
    data : np.ndarray
            the source data array

    indices : np.ndarray
            the indices array created by get_weights

    weights : np.ndarray
            the weights array created by get_weights

    Returns
    -------
    np.ndarray
            the interpolated array
    """
    # create the result array
    result = np.empty(indices.shape[0:2])

    # loop over all target levels
    for target in range(indices.shape[0]):
        # loop over all cells
        for cell in range(indices.shape[1]):
            # if no interpolation is possible, write a nan value
            result[target, cell] = 0
            valid = False
            for val in range(2):
                if weights[target, cell, val] > 0.0:
                    result[target, cell] += data[indices[target, cell, val], cell] * weights[target, cell, val]
                    valid = True
            if not valid:
                result[target, cell] = np.nan
    return result


@jit(nopython=True)
def get_weights(src_p, dst_p):
    """
    calculate indices and weights for each column

    Parameters
    ----------
    src_p : np.ndarray
            array with source pressure grid (lev, cell)

    dst_p : np.ndarray
            array with destination pressure grid (lev)

    Returns
    -------
    indices, weights : tuple
            tuple of arrays with shape (target_levels, cells, 2)
    """
    # arrays for the indices of the neighbouring grid points and the respective weights
    indices = np.empty((dst_p.shape[0], src_p.shape[1], 2), dtype=np.int64)
    weights = np.empty((dst_p.shape[0], src_p.shape[1], 2), dtype=np.float64)

    # loop over all target values
    for target in range(dst_p.shape[0]):
        # loop over all horizontal points
        for cell in range(src_p.shape[1]):
            # find the largest value smaller or equal than the target value and the
            # smallest values larger than the target value
            largest_smaller = -1
            smallerst_larger = -1
            for lev in range(src_p.shape[0]):
                if src_p[lev, cell] <= dst_p[target]:
                    if largest_smaller == -1 or src_p[largest_smaller, cell] < src_p[lev, cell]:
                        largest_smaller = lev
                if src_p[lev, cell] > dst_p[target]:
                    if smallerst_larger == -1 or src_p[smallerst_larger, cell] > src_p[lev, cell]:
                        smallerst_larger = lev

            # store the indices
            indices[target, cell, 0] = largest_smaller
            indices[target, cell, 1] = smallerst_larger

            # use the indices of the neighbour to calculate the weights
            if smallerst_larger != -1 and src_p[smallerst_larger, cell] == dst_p[target]:
                weights[target, cell, 0] = 1.0
                weights[target, cell, 1] = 0.0
            elif largest_smaller == -1 or smallerst_larger == -1:
                weights[target, cell, :] = 0.0
            else:
                weights[target, cell, 0] = 1.0 - (dst_p[target] - src_p[largest_smaller, cell]) / (src_p[smallerst_larger, cell] - src_p[largest_smaller, cell])
                weights[target, cell, 1] = 1.0 - (src_p[smallerst_larger, cell]  - dst_p[target]) / (src_p[smallerst_larger, cell] - src_p[largest_smaller, cell])

    return indices, weights



"""
functions used to create dask-clusters automatically based on the environment a script is executed in.
"""
import os
import sys
import dask
import distributed
import multiprocessing
from .tempdir import TempDir
from .batchjob import get_batch_job, _get_num_available_procs
import atexit
import logging
from time import sleep

# storage the batchjob object
batchjob_object = None

# adapt some settings for dask
from distributed.config import config
config["connect-timeout"] = "30"        # increase the connect-timeout from 3 to 10s


def init_cluster(ntasks=None, extend=False):
    """
    Create a Dask.distributed cluster and return the client object. The type of the cluster is automatically selected
    based on the environment of the script. Inside of a SLURM job, a distributed Cluster is created. All allocated
    resources are used for this purpose. Without a job scheduler like SLURM, a LocalCluster is created.

    Parameters
    ----------
    ntasks : int
            the number of tasks (threads or processes) to start.
    extend : bool
            launch workers in a separate slurm jobs or not
    Returns
    -------
    distributed.Client
            a client object usable to submit computations to the new cluster.
    """
    
    # In case of requesting an extension of the available resources through asking for more workers through SlurmCluster
    # check that the sbatch command its present in the system and call init_slurm_cluster()
    logging.debug("Starting cluster")
    if extend == True and check_sbatch_availability():
        job_id = os.getenv("SLURM_JOB_ID")
        if job_id is None:
            logging.info("Launching new workers through SLURM.")
        else:
            logging.info("Launching new workers through SLURM even do we already are inside a SLURM job with ID %s" %
                         job_id)

        return init_slurm_cluster(nodes=ntasks)

    # create a temporal directory for the work log files
    tmpdir = TempDir(cleanup=False)
    
    # figure out which type of cluster to create
    global batchjob_object
    batchjob_object = get_batch_job(local_dir=tmpdir.getpath(), ntasks=ntasks)

    # start the distributed cluster prepared by the command above
    batchjob_object.start()

    return batchjob_object.get_client()


def init_slurm_cluster(nodes=1, tmp_dir="/dev/shm/"):
    """
    # Submiting DASK workers to a Slurm cluster. Need to merge it with the init_cluster in enstools.core

           Parameters
    ----------
    nodes : int
            number of nodes
    """
    from dask.distributed import Client
    from dask_jobqueue import SLURMCluster

    # Define the kind of jobs that will be launched to the cluster
    # This will apply for each one of the different jobs sent
    cluster = SLURMCluster(
        cores=12,
        memory="24 GB",
        queue="cluster",
        local_directory=tmp_dir,
        #silence_logs="debug",
    )
    # Start workers
    cluster.scale(jobs=nodes)
    client = Client(cluster)
    logging.info("You can follow the dashboard in the following link:\n%s" % client.dashboard_link)
    # client.wait_for_workers(nodes)
    return client    


def get_num_available_procs():
    """
    Get the number of processes available for parallel execution of tasks. If a distributed cluster was started before,
    then the number of workers within this cluster is returned. Otherwise the number of physical processors on the local
    computer. If OMP_NUM_THREADS is defined, it's value will be used!

    Returns
    -------
    int :
            number of processors.
    """
    if batchjob_object is not None:
        return batchjob_object.ntasks
    else:
        return _get_num_available_procs()


def get_client_and_worker():
    """
    This function can be used the get dask distributed client and worker objects.

    Returns
    -------
    tuple :
            (None, None): when called without dask cluster running
            (client, None) when called on a non-worker process
            (client, worker) when called on a worker process
    """
    try:
        client = distributed.get_client()
        logging.debug("get_client_and_worker: client object found!")
    except ValueError:
        client = None
        logging.debug("get_client_and_worker: not running in a dask cluster!")
    if client is not None:
        try:
            worker = distributed.get_worker()
            logging.debug("get_client_and_worker: worker object found!")
        except ValueError:
            worker = None
            logging.debug("get_client_and_worker: not running inside of a worker process!")
    else:
        worker = None
    return client, worker


def all_workers_are_local(client):
    """
    Use the client the get information about the workers.

    Returns
    -------
    bool :
            True is all workers are running on local host
    """
    workers = list(client.scheduler_info()['workers'])
    for worker in workers:
        if not worker.startswith("tcp://127.0.0.1:"):
            return False
    return True


class RoundRobinWorkerIterator():
    def __init__(self, client):
        """
        an iterator that iterates over and over over all workers

        Parameters
        ----------
        client : distributed.client
                the client object of which the worker should be utilised.
        """
        workers = list(client.scheduler_info()['workers'])
        self.workers = list(map(lambda x: tuple(x.replace("tcp://", "").rsplit(":", 1)), workers))
        self.index = 0

    def __iter__(self):
        return self

    def next(self):
        next = self.workers[self.index]  # type: tuple
        self.index += 1
        if self.index == len(self.workers):
            self.index = 0
        return next

    
def check_sbatch_availability():
    """
    Function that checks that sbatch command can be reached in the system.
    It launches the sbatch version command and checks that the return code its 0.
    """
    from subprocess import run, PIPE, CalledProcessError
    
    command = "sbatch --version"
    arguments = command.split()
    result = run(arguments, stdout=PIPE)
    try:
        result.check_returncode()
        return True
    except CalledProcessError:
        logging.warning("Sbatch its not available, won't start an additional cluster.")
        return False


from enstools.core import check_arguments
from .calc_das import threshold_data
from .match_pyramid_ic import match_pyramid
import numpy as np


@check_arguments(shape={"obs": (0, 0), "fct": (0, 0)})
def das(obs, fct, factor=4, threshold=None):
    r"""
    calculate DAS - Displacement and amplitude score for observation and forecast

    based on IDL programm Version 2.0 20090918 C. Keil (DLR) , G.C. Craig (DLR), H. Mannstein (DLR)
    and python scripts from Bettina Richter (Meteorologisches Institut - LMU Muenchen)

    DAS is a displacement-based error measure and includes
    two components: displacement and amplitude error, where

    * displacement is scaled with maximum displacement c1
    * amplitude error is scaled with 'climatological' observed mean c2

    *Philosophy*: complete miss equals 100% amplitude error
    The smaller DAS the better the forecast quality.

    .. math:: das = dis_{nor} + lse_{nor}

    Here both components have equal weight. This is a slight modification
    to the application of FQM on satellite imagery taking the maximum of
    both components published in the reference publication [1]_.


    Parameters
    ----------
    obs : xarray.DataArray or np.ndarray
            2d array with observations
    fct : xarray.DataArray or np.ndarray
            2d array with forecasts
    factor: int
            sub-sampling factor or topmost pyramid level (default: 4)
    threshold : float
            threshold of data (default: `None`)

    Returns
    -------
    dict
            The score itself and all related values:

            das:
                the DAS score
            dis:
                the displacement component
            amp:
                the amplitude component
            rms_obs:
                root mean square of observation
            morph_o:
                morphed fct to obs
            morph_f:
                morphed obs to fct
            xdis_o:
                x-displacement of fct morphed to obs
            ydis_o:
                y-displacement of fct morphed to obs
            xdis_f:
                x-displacement of obs morphed to fct
            ydis_f:
                y-displacement of obs morphed to fct
            lse_o:
                least square error of fct morphed to obs
            lse_f:
                least square error of fct morphed to obs


    References
    ----------
    .. [1] Keil C, Craig G.C., 2009: A displacement and amplitude score
       employing an optical flow technique. WAF, :doi:`10.1175/2009WAF2222247.1`

    .. [2] Keil C, Craig G.C., 2007: A displacement-based error measure applied
       in a regional ensemble forecasting system. MWR 135, 3248-3259, :doi:`10.1175/MWR3457.1`
    """

    # apply threshold
    if threshold is not None:
        obs = threshold_data(obs, threshold)
        fct = threshold_data(fct, threshold)

    # morph fct to obs, obs-space
    morph_o, xdis_o, ydis_o, lse_o = match_pyramid(fct, obs, factor=factor)
    # morph obs to fct, fct-space
    morph_f, xdis_f, ydis_f, lse_f = match_pyramid(obs, fct, factor=factor)

    # calculate DAS - displacement and amplitude score
    dis_max = np.sqrt(2.) * 2. ** (factor + 2.)
    x1 = int(2 * 2 ** factor)
    y2, x2 = [int(s - x1) for s in np.shape(obs)]
    das, dis, amp, rms_obs = calc_das.calc_das(obs, fct, xdis_o, ydis_o, lse_o, xdis_f, ydis_f, lse_f, dis_max,
                                               threshold=threshold, x1=x1, y1=x1, x2=x2, y2=y2)

    # construct the result
    result = {"das": das, "dis": dis, "amp": amp, "rms_obs": rms_obs, "morph_o": morph_o, "morph_f": morph_f,
              "xdis_o": xdis_o, "xdis_f": xdis_f, "ydis_o": ydis_o, "ydis_f": ydis_f, "lse_o": lse_o, "lse_f": lse_f}
    return result



import numpy as np
import xarray
from numpy.ma.core import default_fill_value
from scipy import ndimage
from enstools.core import check_arguments
from enstools.misc import count_ge
from enstools.core.parallelisation import apply_chunkwise


@check_arguments(units={"pr": "kg m-2 s-1",
                        "cape": "J kg-1",
                        "return_value": "hour"},
                 shape={"pr": "cape"})
def convective_adjustment_time_scale(pr, cape, th=1.0, fraction_above_th=0.0015):
    """
    Calculate the convective adjustment time scale from precipitation and CAPE as described in [1]_. A gaussian filter
    is applied to the input data if at least one dimension has more then 30 grid points.

    Parameters
    ----------
    pr:     xarray.DataArray
            Hourly or longer accumulated precipitation converted to [kg m-2 s-1].

    cape:   xarray.DataArray
            The CAPE of mean surface layer parcel [J kg-1].
            For example, mean of start and end value of the period of the accumulation.

    th:     float
            threshold for the precipitation rate above which the calculation should be performed [kg m-2 h-1]. Grid
            points with smaller precipitation values will contain missing values. Default: 1

    fraction_above_th: float
            fraction of grid points that must exceed the threshold defined in `th`. Default: 0.0015 (e.g. 15 of 10.000
            grid points).

    Returns
    -------
    tau:    masked_array or xarray (depending on type of pr)
            the convective adjustment time scale [h]

    Examples
    --------
    >>> if np.__version__ > "1.13.9": np.set_printoptions(legacy="1.13")            # numpy version independent printing
    >>> cape = xarray.DataArray([500.0, 290.44], attrs={"units": "J kg-1"})
    >>> pr   = xarray.DataArray([0.0, 2.0], attrs={"units": "kg m-2 hour-1"})
    >>> np.round(convective_adjustment_time_scale(pr, cape).compute(), 4)                            # doctest:+ELLIPSIS
    <xarray.DataArray 'tauc-...' (dim_0: 2)>
    array([ nan,   1.])
    Dimensions without coordinates: dim_0

    with not enough values above the defined threshold:
    
    >>> np.round(convective_adjustment_time_scale(pr, cape, fraction_above_th=0.6).compute(), 4)     # doctest:+ELLIPSIS
    <xarray.DataArray 'tauc-...' (dim_0: 2)>
    array([ nan,  nan])
    Dimensions without coordinates: dim_0

    References
    ----------
    .. [1]  Keil, C., Heinlein, F. and Craig, G. C. (2014), The convective adjustment time-scale as indicator of
            predictability of convective precipitation. Q.J.R. Meteorol. Soc., 140: 480-490. doi:10.1002/qj.2143
    """

    # TODO: tauc calculation is not chunkwise but something like layer wise
    @apply_chunkwise
    def tauc(pr, cape, th):
        # create a result array filled with the default fill value for the data type of pr
        fill_value = np.nan
        result = np.full_like(pr, fill_value=fill_value)

        # count values above threshold
        n_above_th = count_ge(pr, th / 3600.0)
        if n_above_th < pr.size * fraction_above_th:
            result = np.ma.masked_equal(result, fill_value)
            return result

        # Gaussian filtering
        sig = 10.  # Gaussian goes to zero 3*sig grid points from centre
        if max(pr.shape) > 3*sig:
            cape_filtered = ndimage.filters.gaussian_filter(cape, sig, mode='reflect')
            pr_filtered = ndimage.filters.gaussian_filter(pr, sig, mode='reflect')
        else:
            cape_filtered = cape
            pr_filtered = pr

        # perform the actual calculation
        ind = np.where(pr > th / 3600.0)
        result[ind] = 1.91281e-06 * cape_filtered[ind] / pr_filtered[ind]
        result = np.ma.masked_equal(result, fill_value)
        return result
    result = tauc(pr, cape, th)

    # convert the result to xarray if the input type is also xarray
    if type(pr) == xarray.DataArray:
        result = xarray.DataArray(result, coords=pr.coords, dims=pr.dims, attrs={"units": "hour"})
    return result



import logging
import os
import pandas
from datetime import datetime
from enstools.misc import download, bytes2human, concat
from urllib.error import HTTPError
from urllib.request import urlopen
from enstools.core import get_cache_dir


class DWDContent:
    # create a new temporal directory to store content.log and opendata_dwd_content.pkl
    cache_path = get_cache_dir()

    def __init__(self, refresh_content=False):
        def create_dataframe(logdata):
            """
            Creates a pandas.DataFrame from the given logdata from https://opendata.dwd.de/weather/nwp
            and sets the following attributes:
                file: str
                    Name of the url of the file.
                size: str
                    The size of the file.
                time: np.datetime64
                    The creation time on the server.
                model: str
                    The forecast model, e.g. "icon".
                file_type: str
                    The format of the file ("grib" or "json").
                init_time: int
                    The time of the initialisation of the forecast for the given file.
                variable: str
                    The short name of the variable of the file.
                filename: str
                    The name of the file.
                level_type: str
                    The type of the level of the file.
                forecast_hour: int
                    Hours since the initalization of the forecast.

            Parameters
            ----------
            logdata: str
                The name of the content logfile.

            Returns
            -------
            content: pandas.DataFrame
                The DataFrame object with the given attributes.

            """
            logging.info("Creating content database with {}".format(logdata))
            content = pandas.read_csv(logdata, delimiter="|", header=None, names=["file", "size", "time"])
            content["time"] = pandas.to_datetime(content["time"], format="%Y-%m-%d %H:%M:%S")
            content["size"] = content["size"].astype(int)

            content = content[~content.file.str.contains("snow4")]
            content = content[~content.file.str.contains("content.log")]

            # fix for json data, no json support now:
            content = content[~content.file.str.contains("json")]

            content["model"] = content["file"].apply(lambda x: x.split("/")[1])
            content["file_type"] = content["file"].apply(lambda x: x.split("/")[2])
            content["init_time"] = content["file"].apply(lambda x: x.split("/")[3])
            content["init_time"] = content["init_time"].astype(int)
            content["variable"] = content["file"].apply(lambda x: x.split("/")[4])
            content["filename"] = content["file"].apply(lambda x: x.split("/")[5])
            content["file"] = content["file"].apply(lambda x: "https://opendata.dwd.de/weather/nwp" + x[1:])

            content["abs_init_time"] = content["filename"].apply(lambda x: x.split("_")[4])
            content["level_type"] = content["filename"].apply(lambda x: x.split("_")[3])
            content["grid_type"] = content["filename"].apply(lambda x: x.split("_")[2])
            content["level_type"] = content["level_type"].apply(lambda x: x[:-6])

            content["forecast_hour"] = content["filename"].apply(lambda x: x.split("_")[5])
            content.loc[content["level_type"] == "time-inv", ["forecast_hour"]] = "0"
            content["forecast_hour"] = content["forecast_hour"].astype(int)

            content["level"] = "0"
            content.loc[content["level_type"] == "single", ["level"]] = "0"
            content.loc[content["level_type"] == "pressure", ["level"]] = content[content.level_type == "pressure"][
                "filename"].apply(lambda x: x.split("_")[6])
            content.loc[content["level_type"] == "model", ["level"]] = content[content.level_type == "model"][
                "filename"].apply(lambda x: x.split("_")[6])
            content["level"] = content["level"].astype(int)
            content.to_pickle(self.content_pkl)

            return content
        self.content_pkl = os.path.join(DWDContent.cache_path, "opendata_dwd_content_nwp.pkl")
        if os.path.exists(self.content_pkl) and not refresh_content:
            logging.info(f"Reading content database from {self.content_pkl}")
            content_old = pandas.read_pickle(self.content_pkl)
            self.content = content_old

        else:
            if refresh_content:
                logging.info("Refreshing content database")
            else:
                logging.info("Initializing content database")
            self.content_log_path = os.path.join(DWDContent.cache_path, "content_nwp.log.bz2")
            if os.path.exists(self.content_log_path):
                os.remove(self.content_log_path)

            download("https://opendata.dwd.de/weather/nwp/content.log.bz2",
                     destination=self.content_log_path, uncompress=False)
            self.content = create_dataframe(self.content_log_path)

    def refresh_content(self):
        """
        Initializes the DWDContent object again.
        Downloads the actual content.log from the server and creates a new DataFrame
        """
        self.__init__(refresh_content=True)

    def get_models(self):
        """
        Gives the available models of the data server.
        Returns
        -------
        avail_models: list
            List of Strings with the available models

        """
        content = self.content
        avail_models = content["model"].drop_duplicates().values.tolist()
        avail_models.sort()
        return avail_models

    def get_grid_types(self, model=None):
        """
        Gives the available grid_types for a specific model
        Parameters
        ----------
        model: str
            The model of the forecast simulation.

        Returns
        -------

        """
        content = self.content
        grid_types = content[content["model"] == model]["grid_type"].drop_duplicates().values.tolist()
        return grid_types

    def check_grid_type(self, model=None, grid_type=None):
        """
        Checks if the given grid_type is available and if grid_type=None check if the available type is unambigious.

        Parameters
        ----------
        model: str
            The model of the forecast.
        grid_type: str
            The geo grid type.

        Returns
        -------
        grid_type: str
            if available and unambigious the right grid type will be returned

        """
        avail_grid_types = self.get_grid_types(model=model)
        if grid_type is None:
            if len(avail_grid_types) == 1:
                grid_type = avail_grid_types[0]
            else:
                raise ValueError("You have to choose one of the grid_types: {}".format(avail_grid_types))
        else:
            if grid_type not in avail_grid_types:
                raise ValueError("Grid type {} not available for model {}. available grid_type: {}"
                                 .format(grid_type, model, avail_grid_types))
        return grid_type

    def check_level_type(self, model=None, grid_type=None, init_time=None, variable=None, level_type=None):
        """
        Checks if the given level_type is available for the variable in the given model. If no level_type is given,
        and only one is available, then the one will be returned.

        Parameters
        ----------
        model: str
            The model of the forecast.
        grid_type: str
            The geo grid type.
        init_time: int
            The initialization time.
        variable: list or tuple
            The variable list
        level_type

        Returns
        -------

        """
        avail_level_types = []
        for var in variable:
            alevlist = self.get_avail_level_types(model=model, grid_type=grid_type, init_time=init_time, variable=var)
            # No duplicate values:
            for alev in alevlist:
                if alev not in avail_level_types:
                    avail_level_types.append(alev)
        if level_type is None:
            if len(avail_level_types) == 1:
                level_type = avail_level_types[0]

            else:
                raise ValueError("You have to choose one of the level_types: {}".format(avail_level_types))
        else:
            if level_type not in avail_level_types and len(avail_level_types) > 0:
                raise ValueError("Level type {} not available for model {}, variable {}. Available level_type: {}"
                                 .format(level_type, model, variable, avail_level_types))
        return level_type

    def get_avail_init_times(self, model=None, grid_type=None):
        """
        Gives the available initialization times for a given forecast model.

        Parameters
        ----------
        model: str
            The model for which the available initialization times want to be known.
        grid_type: str
            The type of the geo grid.

        Returns
        -------
        avail_init_times: list
            A list of integers of the available initialization times

        """
        grid_type = self.check_grid_type(model=model, grid_type=grid_type)
        content = self.content
        avail_init_times = content[(content["model"] == model)
                                   & (content["grid_type"] == grid_type)]["init_time"].drop_duplicates().values.tolist()
        avail_init_times.sort()
        avail_abs_init_times = content[(content["model"] == model)
                                       & (content["grid_type"] == grid_type)][
            "abs_init_time"].drop_duplicates().values.tolist()
        avail_abs_init_times.sort()
        for one_abs in avail_abs_init_times:
            avail_init_times.append(datetime.strptime(one_abs, "%Y%m%d%H"))
        return avail_init_times

    def get_avail_vars(self, model=None, grid_type=None, init_time=None):
        """
        Gives the available variables for a given forecast model and initialization time.

        Parameters
        ----------
        model: str
            The model for which the available variables want to be known.
        grid_type: str
            The type of the geo grid.
        init_time: int
            The initialization time for which the available variables want to be known.

        Returns
        -------
        avail_vars: list
            The sorted list of strings of the available variables.

        """
        grid_type = self.check_grid_type(model=model, grid_type=grid_type)
        content = self.content
        avail_vars = content[(content["model"] == model)
                             & (content["grid_type"] == grid_type)
                             & self.__init_time_selection(init_time)]["variable"].drop_duplicates().values.tolist()
        avail_vars.sort()
        return avail_vars

    def get_avail_level_types(self, model=None, grid_type=None, init_time=None, variable=None):
        """
        Gives the available level types for a given forecast model, initialization time and variable.
        Parameters
        ----------
        model: str
            The model for which the available level types want to be known.
        grid_type: str
            The type of the geo grid.
        init_time: int
            The initialization time for which the available level types want to be known.
        variable: str
            The variable for which the available level types want to be known

        Returns
        -------
        avail_level_types: list
            The sorted list of strings of the available level types.

        """
        grid_type = self.check_grid_type(model=model, grid_type=grid_type)
        content = self.content
        avail_level_types = content[(content["model"] == model)
                                    & (content["grid_type"] == grid_type)
                                    & self.__init_time_selection(init_time)
                                    & (content["variable"] == variable)]["level_type"].drop_duplicates().values.tolist()
        avail_level_types.sort()
        return avail_level_types

    def get_avail_forecast_hours(self, model=None, grid_type=None, init_time=None, variable=None, level_type=None):
        """
        Gives the available  forecast hours since initialization for a given forecast model, initialization time,
        variable and level type.

        Parameters
        ----------
        model: str
            The model for which the available forecast hours want to be known.
        grid_type: str
            The type of the geo grid.
        init_time: int
            The initialization time for which the available forecast hours want to be known.
        variable: str
            The variable for which the available forecast hours want to be known.
        level_type: str
            The type of the level for which the available forecast hours want to be known.

        Returns
        -------
        avail_forecast_hours: list
            The available hours of the forecast data since the initilization of the forecast.

        """
        grid_type = self.check_grid_type(model=model, grid_type=grid_type)
        content = self.content
        avail_forecast_times = content[(content["model"] == model)
                                       & (content["grid_type"] == grid_type)
                                       & self.__init_time_selection(init_time)
                                       & (content["variable"] == variable)
                                       & (content["level_type"] == level_type)]["forecast_hour"] \
            .drop_duplicates().values.tolist()
        avail_forecast_times.sort()
        return avail_forecast_times

    def get_avail_levels(self, model=None, grid_type=None, init_time=None, variable=None, level_type=None):
        """
        Gives the available levels since initialization for a given forecast model, initialization time,
        variable and level type. If the level of the variable is not pressure or model 0 will be returned.

        Parameters
        ----------
        model: str
            The model for which the available levels want to be known.
        grid_type: str
            The type of the geo grid.
        init_time: int
            The initialization time for which the available levels want to be known.
        variable: str
            The variable for which the available forecast hours want to be known.
        level_type: str
            The type of the level for which the available levels want to be known.

        Returns
        -------
        avail_levels: list
            A list of integers of the available levels.

        """
        grid_type = self.check_grid_type(model=model, grid_type=grid_type)
        content = self.content
        avail_levels = content[(content["model"] == model)
                               & (content["grid_type"] == grid_type)
                               & self.__init_time_selection(init_time)
                               & (content["variable"] == variable)
                               & (content["level_type"] == level_type)]["level"].drop_duplicates().values.tolist()
        avail_levels.sort()
        return avail_levels

    def __init_time_selection(self, init_time):
        """
        this function is internally used for indexing and selecting a specific init time.

        Parameters
        ----------
        init_time:  int or datetime
                    if an integer is given, the column inti_time is used. For da datetime object
                    the column abs_init_time is used.
        Returns
        -------
        Series:
                    pandas series object used for indexing.
        """
        if type(init_time) == int:
            return self.content["init_time"] == init_time
        elif isinstance(init_time, datetime):
            return self.content["abs_init_time"] == init_time.strftime("%Y%m%d%H")
        else:
            raise ValueError(f"invalid data type for init_time: {type(init_time)}")

    def __get_uniq_content_line(self, model=None, grid_type=None, init_time=None, variable=None,
                                level_type=None, forecast_hour=None, level=None):
        """
        Use the given parameters to find ONE entry of the content.

        Parameters
        ----------
        model: str
            The model of the file for which the url wants to be known.
        grid_type: str
            The type of the geo grid.
        init_time: int
            The initialization time of the forecast of the file for which the url wants to be known.
        variable: str
            The variable of the file for which the url wants to be known.
        level_type: str
            The type of level of the file for which the url wants to be known.
        forecast_hour: int
            The hours since the initialization of the forecast of the file for which the url wants to be known.
        level: int
            The level of the file for which the url wants to be known.

        Returns
        -------
        pandas.DataFrame
        """
        grid_type = self.check_grid_type(model=model, grid_type=grid_type)
        content = self.content
        result = content[(content["model"] == model)
                         & self.__init_time_selection(init_time)
                         & (content["variable"] == variable)
                         & (content["level_type"] == level_type)
                         & (content["forecast_hour"] == forecast_hour)
                         & (content["level"] == level)
                         & (content["grid_type"] == grid_type)]

        # raise exception if nothing was found.
        if len(result) == 0:
            raise KeyError("No entry not found in content for (model: {}, grid_type: {}, init_time: {}, variable: {}, "
                           "level_type: {}, forecast_hour: {}, level: {})"
                           .format(model, grid_type, init_time, variable, level_type, forecast_hour, level))
        elif len(result) > 1:
            logging.warning("{} entries found in content, only one expected for (model: {}, grid_type: {}, "
                            "init_time: {}, variable: {}, level_type: {}, forecast_hour: {}, level: {}), "
                            "taking the newest!"
                            .format(len(result), model, grid_type, init_time, variable, level_type, forecast_hour,
                                    level))
            result = result.sort_values(by="time", ascending=False).iloc[0]
        return result

    def get_url(self, model=None, grid_type=None, init_time=None, variable=None,
                level_type=None, forecast_hour=None, level=None):
        """
        Gives the url of the file on the https://opendata.dwd.de/weather/nwp server.

        Parameters
        ----------
        model: str
            The model of the file for which the url wants to be known.
        grid_type: str
            The type of the geo grid.
        init_time: int
            The initialization time of the forecast of the file for which the url wants to be known.
        variable: str
            The variable of the file for which the url wants to be known.
        level_type: str
            The type of level of the file for which the url wants to be known.
        forecast_hour: int
            The hours since the initialization of the forecast of the file for which the url wants to be known.
        level: int
            The level of the file for which the url wants to be known.

        Returns
        -------
        url: str
            The url adress of the file.
        """
        url = self.__get_uniq_content_line(model=model, grid_type=grid_type, init_time=init_time, variable=variable,
                                           level_type=level_type, forecast_hour=forecast_hour, level=level)
        return url["file"].values[0]

    def get_filename(self, model=None, grid_type=None, init_time=None, variable=None,
                     level_type=None, forecast_hour=None, level=None):
        """
         Gives the filename of the file on the https://opendata.dwd.de/weather/nwp server.

        Parameters
        ----------
        model: str
            The model of the file for which the filename wants to be known.
        grid_type: str
            The type of the geo grid.
        init_time: int
            The initialization time of the forecast of the file for which the filename wants to be known.
        variable: str
            The variable of the file for which the filename wants to be known.
        level_type: str
            The type of level of the file for which the filename wants to be known.
        forecast_hour: int
            The hours since the initialization of the forecast of the file for which the filename wants to be known.
        level: int
            The level of the file for which the filename wants to be known.

        Returns
        -------
        url: str
            The filename of the file.
        """
        url = self.__get_uniq_content_line(model=model, grid_type=grid_type, init_time=init_time, variable=variable,
                                           level_type=level_type, forecast_hour=forecast_hour, level=level)
        return url["filename"].values[0]

    def check_url_available(self, model=None, grid_type=None, init_time=None, variable=None,
                            level_type=None, forecast_hour=None, level=None):
        grid_type = self.check_grid_type(model=model, grid_type=grid_type)
        try:
            http_code = urlopen(self.get_url(model=model, grid_type=grid_type, init_time=init_time, variable=variable,
                                             level_type=level_type, forecast_hour=forecast_hour, level=level)).getcode()
            if http_code == 200:
                url_available = True
        except HTTPError:
            url_available = False
        return url_available

    def get_file_size(self, model=None, grid_type=None, init_time=None, variable=None,
                      level_type=None, forecast_hour=None, level=None):
        url = self.__get_uniq_content_line(model=model, grid_type=grid_type, init_time=init_time, variable=variable,
                                           level_type=level_type, forecast_hour=forecast_hour, level=level)
        return url["size"].values[0]

    def get_size_of_download(self, model=None, grid_type=None, init_time=None,
                             variable=None, level_type=None, forecast_hour=None, levels=None):
        grid_type = self.check_grid_type(model=model, grid_type=grid_type)
        total_size = 0
        for var in variable:
            for hour in forecast_hour:
                for lev in levels:
                    total_size += self.get_file_size(model=model, grid_type=grid_type, init_time=init_time,
                                                     variable=var, level_type=level_type, forecast_hour=hour, level=lev)
        return total_size

    def check_parameters(self, model=None, grid_type=None, init_time=None, variable=None,
                         level_type=None, forecast_hour=None, levels=None):
        """
        Checks if there are all files available for the given parameters.
        If not, the DWDContent object will be refreshed.
        If one file is not available for the given parameters, a detailed error will be thrown.

        Parameters
        ----------
        model:str
            The model of the file.
        grid_type: str
            The type of the geo grid.
        init_time: int
            The initialization time of the file.
        variable: list or str
            The variable of the file.
        level_type: str
            The type of level of the file.
        forecast_hour: list or int
            The hours of the forecast since the initialization of the simulation.
        levels: list or int
            The levels.

        Returns
        -------

        """
        grid_type = self.check_grid_type(model=model, grid_type=grid_type)

        params_available = True
        if model not in self.get_models():
            params_available = False

        if init_time not in self.get_avail_init_times(model=model, grid_type=grid_type):
            params_available = False

        for var in variable:
            avail_vars = self.get_avail_vars(model=model, grid_type=grid_type, init_time=init_time)
            if var not in avail_vars:
                params_available = False
                break

            avail_level_types = self.get_avail_level_types(model=model, grid_type=grid_type,
                                                           init_time=init_time, variable=var)
            if level_type not in avail_level_types:
                params_available = False
                break

            for hour in forecast_hour:
                if hour not in self.get_avail_forecast_hours(model=model, grid_type=grid_type, init_time=init_time,
                                                             variable=var, level_type=level_type):
                    params_available = False
                    break

                for lev in levels:
                    avail_levels = self.get_avail_levels(model=model, grid_type=grid_type, init_time=init_time,
                                                         variable=var, level_type=level_type)

                    if lev not in avail_levels:
                        params_available = False
                        break

        if not params_available:
            logging.warning("Parameters not available")
        else:
            for hour in forecast_hour:
                for var in variable:
                    for lev in levels:
                        if not self.check_url_available(model=model, grid_type=grid_type, init_time=init_time,
                                                        variable=var, level_type=level_type, forecast_hour=hour,
                                                        level=lev):
                            params_available = False
                            break
            if not params_available:
                logging.warning("Database outdated.")

        if not params_available:
            self.refresh_content()

            avail_models = self.get_models()
            if model not in avail_models:
                raise ValueError("The model {} is not available. Possible Values: {}".format(model, avail_models))

            avail_init_times = self.get_avail_init_times(model=model, grid_type=grid_type)
            if init_time not in avail_init_times:
                raise ValueError("The initial time {} is not available. Possible Values: {}"
                                 .format(init_time, avail_init_times))

            for var in variable:
                avail_vars = self.get_avail_vars(model=model, grid_type=grid_type, init_time=init_time)
                if var not in avail_vars:
                    raise ValueError("The variable {} is not available for the {} model "
                                     "and the init_time {}. Available variables: {}"
                                     .format(var, model, init_time, avail_vars))

                avail_level_types = self.get_avail_level_types(model=model, grid_type=grid_type,
                                                               init_time=init_time, variable=var)
                if level_type not in avail_level_types:
                    raise ValueError("The level type {} is not available for the variable {}. Available types: {}"
                                     .format(level_type, var, avail_level_types))
                avail_forecast_hours = self.get_avail_forecast_hours(model=model, grid_type=grid_type,
                                                                     init_time=init_time, variable=var,
                                                                     level_type=level_type)
                for hour in forecast_hour:

                    if hour not in avail_forecast_hours:
                        raise ValueError(
                            "The forecast hour {} is not available for the variable {}. Possible values: {}"
                                .format(hour, var, avail_forecast_hours))
                    for lev in levels:
                        avail_levels = self.get_avail_levels(model=model, grid_type=grid_type, init_time=init_time,
                                                             variable=var, level_type=level_type)
                        if lev not in avail_levels:
                            raise ValueError("The level {} is not available for the variable {} and the level type {}."
                                             .format(lev, var, level_type)
                                             + " Possible Values: {}".format(avail_levels))

    def get_merge_dataset_name(self, model=None, variable=None,
                               level_type=None, init_time=None, forecast_hour=None):
        # find the absolute init time for the given init_time, use only the first variable
        abs_init_time = self.content[(self.content["model"] == model) &
                                     (self.content["level_type"] == level_type) &
                                     (self.content["variable"] == variable[0]) &
                                     self.__init_time_selection(init_time)]["abs_init_time"].drop_duplicates().values[0]

        # construct the actual file name
        merge_name = model + "_" + level_type + "_" + "init" + abs_init_time
        # add forecast time
        # do we have regular spaced data?
        if len(forecast_hour) == forecast_hour[-1] - forecast_hour[0] + 1:
            merge_name += f"+{forecast_hour[0]}h-{forecast_hour[-1]}h"
        else:
            for one_time in forecast_hour:
                merge_name += "+{}h".format(one_time)
        merge_name += "_"
        # add variable names
        for var in variable:
            merge_name = merge_name + var + "+"
        merge_name = merge_name[:-1]

        return merge_name + ".grib2"

    def retrieve(self, service="DWD", model="ICON", eps=None, grid_type=None, variable=None, level_type=None,
                 levels=0, init_time=None, forecast_hour=None, merge_files=False, dest=None):
        """
        Downloads datasets from opendata server. Faster access to the database.
        Parameters
        ----------
        service : str
                name of weather service. Default="DWD".
        model : str
                name of the model. Default="ICON".

        eps : bool
                if True, download ensemble forecast, otherwise download deterministic forecast.

        grid_type: str
            The type of the geo grid.

        variable : list or str
                list of variables to download. Multiple values are allowed.

        level_type : str
                one of "model", "pressure", or "single"

        levels : list or int
                levels to download. Unit depends on `level_type`.

        init_time : int or str

        forecast_hour : list or int
                hours since the initialization of forecast. Multiple values are allowed.

        merge_files : bool
                if true, GRIB files are concatenated to create one file.

        dest : str
                Destination folder for downloaded data. If the files are already available,
                they are not downloaded again.

        Returns
        -------
        list :
                names of downloaded files.
        """

        # Want to download one or more variables?
        if not isinstance(variable, (list, tuple)):
            variable = [variable]
        # Want to download one or more forecast hours?
        if not isinstance(forecast_hour, (list, tuple)):
            forecast_hour = [forecast_hour]
        if not isinstance(levels, (list, tuple)):
            levels = [levels]
        if service.lower() != "dwd":
            raise KeyError("Only DWD server is available")

        if not os.path.exists(dest):
            os.mkdir(dest)
        model = model.lower()
        if grid_type is not None:
            grid_type = grid_type.lower()
        if level_type is not None:
            level_type = level_type.lower()
        variable = [var.lower() for var in variable]

        if model.endswith("eps") and eps is False:
            raise ValueError("{} is a ensemble forecast, but eps was set to False!".format(model))
        elif eps is True and not model.endswith("-eps"):
            model = model + "-eps"

        download_files = []
        download_urls = []

        grid_type = self.check_grid_type(model=model, grid_type=grid_type)
        level_type = self.check_level_type(model=model, grid_type=grid_type, init_time=init_time,
                                           variable=variable, level_type=level_type)

        self.check_parameters(model=model, grid_type=grid_type, init_time=init_time, variable=variable,
                              level_type=level_type, forecast_hour=forecast_hour, levels=levels)
        if merge_files:
            merge_dataset_name = dest + "/" + self.get_merge_dataset_name(model=model, variable=variable,
                                                                          level_type=level_type, init_time=init_time,
                                                                          forecast_hour=forecast_hour)
            if os.path.exists(merge_dataset_name):
                logging.warning("file not downloaded because it is already present: " + merge_dataset_name)
                return [merge_dataset_name]
        total_size = 0
        for var in variable:
            for hour in forecast_hour:
                for lev in levels:
                    content_entry = self.__get_uniq_content_line(model=model, grid_type=grid_type, init_time=init_time,
                                                                 variable=var, level_type=level_type,
                                                                 forecast_hour=hour,
                                                                 level=lev)
                    download_urls.append(content_entry["file"].values[0])
                    download_files.append(content_entry["filename"].values[0])
                    total_size += content_entry["size"].values[0]

        download_files = [dest + "/" + file[:-4] for file in download_files]

        total_size_human = bytes2human(total_size)

        logging.info("Downloading {} files with the total size of {}".format(len(download_files), total_size_human))
        for i in range(len(download_urls)):
            download(download_urls[i], download_files[i] + ".bz2", uncompress=True)

        if merge_files:
            concat(download_files, merge_dataset_name)
            for file in download_files:
                os.remove(file)
            return [merge_dataset_name]
        else:
            return download_files


# store the created content object for later re-use
__content = None


def getDWDContent():
    """
    Creates a DWDContent object and returns it. Multiple calls to this function will return the same object.

    Returns
    -------
    DWDContent
    """
    global __content
    if __content is not None:
        return __content
    else:
        __content = DWDContent()
        return __content



from enstools.core import check_arguments
import numpy as np


@check_arguments(shape={"arr": (0, 0)})
def downsize(arr, fac):
    """
    Reduce resolution of an array by neighbourhood averaging - 2D averaging of fac x fac element

    Parameters
    ----------
    arr : xarray.DataArray or np.ndarray
            array to downsize by neighbourhood averaging

    fac : int
            factor of downsizing, 2D averaging of fac x fac element

    Returns
    -------
    xarray.DataArray or np.ndarray
    """
    row, col = np.shape(arr)
    r_small = int(row // fac)
    c_small = int(col // fac)
    return np.column_stack(np.column_stack(
        arr.reshape((r_small, row // r_small, c_small, col // c_small)))).mean(1).reshape((r_small, c_small))


